import enum
import logging
import crc8
import traceback
import asyncio

from joycontrol.controller_state import ControllerState
from joycontrol.nfc_tag import NFCTag

logger = logging.getLogger(__name__)


###############################################################
## This simulates the MCU in the right joycon/Pro-Controller ##
###############################################################
# WARNING: THIS IS ONE GIANT RACE CONDITION, DON'T DO THINGS FAST
# No I won't fix this, I have had enough of this asyncio b***s***
# DIY or STFU
# This is sufficient to read or write one amiibo when simulating
# a Pro-Controller
# multiple can mess up the internal state
# anything but amiboo is not supported
# TODO:
#     - figure out the NFC-content transfer, currently everything is hardcoded to work wich amiibo
#       see https://github.com/CTCaer/jc_toolkit/blob/5.2.0/jctool/jctool.cpp l2456ff for sugesstions
#     - IR-camera
#     - writing to Amiibo the proper way
#     - verify right joycon
#     - Figure out the UID index in write commands, currently the check is just uncommented...

# These Values are used in set_power, set_config and get_status packets
# But not all of them can appear in every one
# see https://github.com/CTCaer/jc_toolkit/blob/5.2.0/jctool/jctool.cpp l 2359 for set_config
class MCUPowerState(enum.Enum):
    SUSPENDED = 0x00  # set_power
    READY = 0x01  # set_power, set_config, get_status
    READY_UPDATE = 0x02
    CONFIGURED_NFC = 0x04  # set_config, get_status
    # CONFIGURED_IR = 0x05  # TODO: support this
    # CONFIGUERED_UPDATE = 0x06

SET_POWER_VALUES = (
    MCUPowerState.SUSPENDED.value,
    MCUPowerState.READY.value,
    #   MCUPowerState.READY_UPDATE.value,
)

SET_CONFIG_VALUES = (
    MCUPowerState.READY.value,
    MCUPowerState.CONFIGURED_NFC.value,
    #   MCUPowerState.CONFIGURED_IR.value,
)

GET_STATUS_VALUES = (
    MCUPowerState.READY.value,
    #   MCUPowerState.READY_UPDATE.value,
    MCUPowerState.CONFIGURED_NFC.value,
    #   MCUPowerState.CONFIGURED_IR.value
)


def MCU_crc(data):
    if not isinstance(data, bytes):
        data = bytes(data)
    my_hash = crc8.crc8()
    my_hash.update(data)
    # At this point I'm not even mad this works...
    return my_hash.digest()[0]


class NFC_state(enum.Enum):
    NONE = 0x00
    POLL = 0x01
    PENDING_READ = 0x02
    WRITING = 0x03
    AWAITING_WRITE = 0x04
    PROCESSING_WRITE = 0x05
    POLL_AGAIN = 0x09


def pack_message(*args, background=0, checksum=MCU_crc, length=313):
    """
    convinience function that packes
    * hex strings
    * byte lists
    * integer lists
    * Enums
    * integers
    into a 313 bytes long MCU response
    """
    data = bytearray([background] * length)
    cur = 0
    for arg in args:
        if isinstance(arg, str):
            arg = bytes.fromhex(arg)
        elif isinstance(arg, int):
            arg = bytes([arg])
        elif isinstance(arg, enum.Enum):
            arg = bytes([arg.value])
        else:
            arg = bytes(arg)
        arg_len = len(arg)
        if arg_len + cur > length:
            logger.warning("MCU: too long message packed")
        data[cur:cur + arg_len] = arg
        cur += arg_len
    if checksum:
        data[-1] = checksum(data[0:-1])
    return data


class MicroControllerUnit:
    def __init__(self, controller: ControllerState):

        self.power_state = MCUPowerState.SUSPENDED

        # NOT USED
        # Just a store for the remaining configuration data
        self.configuration = None

        self.nfc_state = NFC_state.NONE
        # Counter for testing state transitions
        self.nfc_counter = 0

        self._last_poll_uid = None

        # after writing we need to "remove" the amiibo
        # sending the detection of a bunch of zeros works
        # (pretty sure it just panics the switch)
        self.remove_amiibo = NFCTag(data=bytes(540))
        # weather or not the next n POLLs without a tag should assume the remove_amiibo
        self._pending_active_remove = 0

        # remove the tag from the controller after a successful read
        # NOT IMPLEMENTED self.remove_nfc_after_read = False
        # remove the tag from the controller after a successful write, the switch demands this
        self.remove_nfc_after_write = True

        # controllerstate to look for nfc-data
        self._controller = controller

        # long messages are split, this keeps track of in and outgoing transfers
        self.seq_no = 0
        self.ack_seq_no = 0
        # self.expect_data = False
        self.received_data = []

        # We are getting 0x11 commands to do something, but cannot answer directly
        # responses have to be passed in regular input reports
        # If there was no command, this is the default report
        self.no_response = pack_message(0xff)
        self.response_queue = []
        # to prevent the queue from becoming too laggy, limit it's size
        # the length after which we start dropping packets
        self.max_response_queue_len = 4

    def _flush_response_queue(self):
        self.response_queue = []

    def _queue_response(self, resp):
        if len(self.response_queue) < self.max_response_queue_len:
            self.response_queue.append(resp)
        else:
            logger.warning("Full queue, dropped outgoing MCU packet")

    # used to queue messages that cannot be dropped, as we don't have any kind of resend-mechanism
    def _force_queue_response(self, resp):
        self.response_queue.append(resp)
        if len(self.response_queue) > self.max_response_queue_len:
            logger.warning("Forced response queue")

    def set_remove_nfc_after_read(self, value):
        pass
        # self.remove_nfc_after_read = value

    def _get_status_data(self, args=None):
        """
        create a status packet to be used when responding to 1101 commands (outside NFC-mode)
        """
        if self.power_state == MCUPowerState.SUSPENDED:
            logger.warning("MCU: status request when disabled")
            return self.no_response
        elif self.power_state.value in GET_STATUS_VALUES:
            return pack_message("0100000008001b", self.power_state)

    def _get_nfc_status_data(self, args):
        """
        Generate a NFC-Status report to be sent back to switch
        This is 40% of all logic in this file, as all we do in NFC-mode is send these responses
        but some (the read/write-tag ones) are hardcoded somewhere else
        @param args: not used
        @return: the status-message
        """
        self.nfc_counter -= 1
        nfc_tag = self._controller.get_nfc()

        # after a write to get out of the screen report the empty amiibo
        if self.nfc_state in (NFC_state.POLL, NFC_state.POLL_AGAIN) and (self.remove_nfc_after_write or not nfc_tag) and self._pending_active_remove > 0:
            nfc_tag = self.remove_amiibo
            self._pending_active_remove -= 1

        if self.nfc_state == NFC_state.PROCESSING_WRITE and self.nfc_counter <= 0:
            self.nfc_state = NFC_state.NONE
        elif self.nfc_state == NFC_state.POLL:
            if nfc_tag and nfc_tag.getUID() == self._last_poll_uid:
                self.nfc_state = NFC_state.POLL_AGAIN
            else:
                self._last_poll_uid = nfc_tag.getUID() if nfc_tag else None
        elif self.nfc_state == NFC_state.POLL_AGAIN:
            if not nfc_tag or nfc_tag.getUID() != self._last_poll_uid:
                self.nfc_state = NFC_state.POLL
                self._last_poll_uid = nfc_tag.getUID() if nfc_tag else None

        if nfc_tag and self.nfc_state != NFC_state.NONE:
            # states POLL, POLL_AGAIN, AWAITING_WRITE, WRITING, PROCESSING_WRITE can include the UID
            out = pack_message("2a0005", self.seq_no, self.ack_seq_no, "0931", self.nfc_state,
                               "0000000101020007", nfc_tag.getUID())
        else:  # seqno and ackseqno should be 0 if we're not doing anything fancy
            out = pack_message("2a000500000931", self.nfc_state)

        return out

    async def process_nfc_write(self, command):
        """
        After all data regarding the write to the tag has been received, this function actually applies the changed
        @param command: the entire write request
        @return: None
        """
        logger.info("MCU: processing nfc write")
        nfc_tag: NFCTag = self._controller.get_nfc()
        if not nfc_tag:
            logger.error("nfc_tag is none, couldn't write")
            return
        if command[1] != 0x07:  # panic wrong UUID length
            logger.error(f"UID length is {command[1]} (not 7), aborting")
            return
        if bytes(command[2:9]) != nfc_tag.getUID():
            logger.error(f"self.nfc_tag.uid and target uid aren't equal, are {bytes(nfc_tag.getUID()).hex()} and {bytes(command[2:9]).hex()}")
            # return # wrong UUID, won't write to wrong UUID
        if nfc_tag.is_mutable():
            nfc_tag.create_backup()
        else:
            nfc_tag.set_mutable(True)

        # write write-lock
        nfc_tag.data[16:20] = command[13:17]

        i = 22
        while i + 1 < len(command):
            addr = command[i] * 4
            leng = command[i + 1]
            data = command[i + 2:i + 2 + leng]
            if addr == 0 or leng == 0:
                break
            nfc_tag.write(addr, data)
            i += 2 + leng

        # remove write lock
        nfc_tag.data[16:20] = command[17:21]
        nfc_tag.save()
        return

    def handle_nfc_subcommand(self, com, data):
        """
        This generates responses for NFC commands and thus implements the entire
        NFC-behaviour
        @param com: the NFC-command (not the 0x02, the byte after that)
        @param data: the remaining data
        """
        if com == 0x04:  # status / response request
            # the switch spams this up to 8 times a frame, there is no way to respond to all
            self._queue_response(self._get_nfc_status_data(data))
        elif com == 0x01:  # start polling, should we queue a nfc_status?
            logger.debug("MCU-NFC: start polling")
            self.nfc_state = NFC_state.POLL
        elif com == 0x06:  # read/write
            logger.debug(f"MCU-NFC Read/write {data[6:13]}")
            nfc_tag = self._controller.get_nfc()
            if nfc_tag:
                # python usually doesn't care about data-types but a list is not an array.... how I hate this crap
                if bytes(data[6:13]) == bytes(7):  # This is the UID, 0 means read anything
                    logger.info("MCU-NFC: reading Tag...")
                    self._flush_response_queue()
                    # Data is sent in 2 packages plus a trailer
                    # the first one contains a lot of fixed header and the UUID
                    # the length and packetnumber is in there somewhere, see
                    # https://github.com/CTCaer/jc_toolkit/blob/5.2.0/jctool/jctool.cpp line 2523ff
                    self._force_queue_response(pack_message(
                        "3a0007010001310200000001020007",
                        nfc_tag.getUID(),
                        "000000007DFDF0793651ABD7466E39C191BABEB856CEEDF1CE44CC75EAFB27094D087AE803003B3C7778860000",
                        nfc_tag.data[0:245]
                    ))
                    # the second one is mostely data, followed by presumably anything (zeroes work)
                    self._force_queue_response(pack_message(
                        "3a000702000927",
                        nfc_tag.data[245:540]
                    ))
                    # the trailer includes the UUID again
                    # this is not actually the trailer, joycons send both packets again but with the first bytes in the
                    # first one 0x2a then this again. But it seems to work without?
                    self._force_queue_response(pack_message(
                        "2a000500000931040000000101020007",
                        nfc_tag.getUID()
                    ))
                # elif bytes(data[6:13]) == nfc_tag.getUID():  # we should check the UID
                else:  # the UID is nonzero, so I a write to that tag follows
                    logger.info(f"MCU-NFC: setup writing tag {data[6:13]}")
                    self._force_queue_response(pack_message(
                        "3a0007010008400200000001020007", nfc_tag.getUID(),  # standard header for write, some bytes differ from read
                        "00000000fdb0c0a434c9bf31690030aaef56444b0f602627366d5a281adc697fde0d6cbc010303000000000000f110ffee"
                        # any guesses are welcome. The end seems like something generic, a magic number?
                    ))
                    self.received_data = []
                    self.nfc_state = NFC_state.AWAITING_WRITE
            else:
                logger.error("had no NFC tag when read/write was initiated")
        # elif com == 0x00: # cancel eveyhting -> exit mode?
        elif com == 0x02:  # stop polling, respond?
            logger.debug("MCU-NFC: stop polling...")
            self.nfc_state = NFC_state.NONE
            self._last_poll_uid = None
        elif com == 0x08:  # write NTAG
            if data[0] == 0 and data[2] == 0x08:  # never seen, single packet as entire payload
                asyncio.ensure_future(self.process_nfc_write(data[4: 4 + data[3]]))
                logger.warning("MCU-NFC write valid but WTF")
                return
            if data[0] <= self.ack_seq_no:  # we already saw this one
                logger.info(f"MCU-NFC write packet repeat {data[0]}")
                pass
            elif data[0] == 1 + self.ack_seq_no:  # next packet in sequence
                self.received_data += data[4: 4 + data[3]]
                self.ack_seq_no += 1
                logger.debug(f"MCU-NFC write packet {self.ack_seq_no}")
            else:  # panic we missed/skipped something
                logger.warning(f"MCU-NFC write unexpected packet, expected {self.ack_seq_no} got {data[0]}, aborting.")
                self.ack_seq_no = 0
            self.nfc_state = NFC_state.WRITING
            self._force_queue_response(self._get_nfc_status_data(data))
            if data[2] == 0x08:  # end of sequence
                self.ack_seq_no = 0
                self.nfc_state = NFC_state.PROCESSING_WRITE
                self.nfc_counter = 4
                self._pending_active_remove = 4 # Dunno, anything > 2 works most likely
                asyncio.ensure_future(self.process_nfc_write(self.received_data))
        else:
            logger.error("unhandled NFC subcommand", com, data)

    # I don't actually know if we are supposed to change the MCU-data based on
    # regular subcommands, but the switch is spamming the status-requests anyway,
    # so sending responses seems to not hurt

    # protocoll-callback
    def entered_31_input_mode(self):
        self._flush_response_queue()
        self.power_state = MCUPowerState.READY
        self._queue_response(self._get_status_data())

    # protocoll-callback
    def set_power_state_cmd(self, power_state):
        logger.debug(f"MCU: Set power state cmd {power_state}")
        self._flush_response_queue()
        if power_state in SET_POWER_VALUES:
            self.power_state = MCUPowerState(power_state)
        else:
            logger.error(f"not implemented power state {power_state}")
            self.power_state = MCUPowerState.READY
        self._queue_response(self._get_status_data())

    # protocoll-callback
    def set_config_cmd(self, config):
        if self.power_state == MCUPowerState.SUSPENDED:
            if config[2] == 0:
                # the switch does this during initial setup, presumably to disable
                # any MCU weirdness
                pass
            else:
                logger.warning("Set MCU Config not in READY mode")
        elif self.power_state == MCUPowerState.READY:
            if config[2] in SET_CONFIG_VALUES:
                self.power_state = MCUPowerState(config[2])
                self.configuration = config
                logger.debug(f"MCU Set configuration {self.power_state} {self.configuration}")
            else:
                self.power_state = MCUPowerState.READY
                logger.error(f"Not implemented configuration written {config[2]} {config}")
            if self.power_state == MCUPowerState.CONFIGURED_NFC:
                # reset all nfc-related parts
                self.nfc_state = NFC_state.NONE
                self.nfc_counter = 0
                self.seq_no = 0
                self.ack_seq_no = 0
            self._queue_response(self._get_status_data())

    # protocoll-callback
    def received_11(self, subcommand, subcommanddata):
        """
        This function handles all 0x11 output-reports.
        @param subcommand: the subcommand as integer
        @param subcommanddata: the remaining data
        @return: None
        """
        if subcommand == 0x01:
            # status request
            self._queue_response(self._get_status_data(subcommanddata))
        elif subcommand == 0x02:
            # NFC command
            if self.power_state != MCUPowerState.CONFIGURED_NFC:
                logger.warning(f"NFC command ({subcommand} {subcommanddata}) outside NFC mode, ignoring")
            else:
                self.handle_nfc_subcommand(subcommanddata[0], subcommanddata[1:])
        else:
            logger.error(f"unknown 0x11 subcommand {subcommand} {subcommanddata}")

    # protocoll hook
    def get_data(self):
        """
        The function returning what is to write into mcu data of tha outgoing 0x31 packet

        usually it is some queued response
        @return: the data
        """
        if len(self.response_queue) > 0:
            return self.response_queue.pop(0)
        else:
            return self.no_response



from time import perf_counter
from json import dumps


DIRECT_INPUT_IDLE_PACKET = {
    # Sticks
    "L_STICK": {
        "PRESSED": False,
        "X_VALUE": 0,
        "Y_VALUE": 0,
        # Keyboard position calculation values
        "LS_UP": False,
        "LS_LEFT": False,
        "LS_RIGHT": False,
        "LS_DOWN": False
    },
    "R_STICK": {
        "PRESSED": False,
        "X_VALUE": 0,
        "Y_VALUE": 0,
        # Keyboard position calculation values
        "RS_UP": False,
        "RS_LEFT": False,
        "RS_RIGHT": False,
        "RS_DOWN": False
    },
    # Dpad
    "DPAD_UP": False,
    "DPAD_LEFT": False,
    "DPAD_RIGHT": False,
    "DPAD_DOWN": False,
    # Triggers
    "L": False,
    "ZL": False,
    "R": False,
    "ZR": False,
    # Joy-Con Specific Buttons
    "JCL_SR": False,
    "JCL_SL": False,
    "JCR_SR": False,
    "JCR_SL": False,
    # Meta buttons
    "PLUS": False,
    "MINUS": False,
    "HOME": False,
    "CAPTURE": False,
    # Buttons
    "Y": False,
    "X": False,
    "B": False,
    "A": False
}


class InputParser():

    # Left Stick calibration values
    LEFT_STICK_CALIBRATION = {
        "center_x": 2159,
        "center_y": 1916,
        # Zeroed Min/Max X and Y
        "min_x": -1466,
        "max_x": 1517,
        "min_y": -1583,
        "max_y": 1465,
    }
    # Right Stick calibration values
    RIGHT_STICK_CALIBRATION = {
        "center_x": 2070,
        "center_y": 2013,
        # Zeroed Min/Max X and Y
        "min_x": -1522,
        "max_x": 1414,
        "min_y": -1531,
        "max_y": 1510,
    }

    def __init__(self, protocol):

        self.protocol = protocol

        # Buffers a list of unparsed macros
        self.macro_buffer = []

        # Keeps track of the entire current
        # list of macro commands.
        self.current_macro = None
        self.current_macro_id = None
        # Keeps track of the macro commands being
        # input over a period of time.
        self.current_macro_commands = None

        # The time length of the current macro
        self.macro_timer_length = 0

        # The start time for the current macro commands
        self.macro_timer_start = 0

        self.controller_input = None

        # Whether or not input has been entered
        # that would close the "Change Grip/Order" menu
        self.exited_grip_order_menu = False

    def buffer_macro(self, macro, macro_id):

        # Doesn't have any info
        if len(macro) < 4:
            return

        self.macro_buffer.append([macro, macro_id])

    def stop_macro(self, macro_id, state=None):

        # Check if the macro is being input currently
        if macro_id == self.current_macro_id:
            # If so, reset the current macro
            self.current_macro = None
            self.current_macro_id = None
            self.current_macro_commands = None
            self.macro_timer_length = 0
            self.macro_timer_start = 0
        else:
            # Check if the macro is still in the buffer
            for i in range(0, len(self.macro_buffer)):
                if macro_id == self.macro_buffer[i][1]:
                    del self.macro_buffer[i]

        # Ensure the stopped macro is added to the finished
        # macros so that any blocking parties listening can
        # continue.
        if state:
            finished = state["finished_macros"]
            finished.append(macro_id)
            state["finished_macros"] = finished

        return

    def clear_macros(self):

        self.current_macro = None
        self.current_macro_id = None
        self.current_macro_commands = None
        self.macro_timer_length = 0
        self.macro_timer_start = 0
        self.macro_buffer = []

        return

    def set_controller_input(self, controller_input):

        self.controller_input = controller_input
    
    def commands_queued(self):
        check = dumps(self.controller_input) != dumps(DIRECT_INPUT_IDLE_PACKET)
        check = check or self.macro_buffer
        check = check or self.current_macro
        check = check or self.current_macro_commands
        return check

    def active_input_queued(self):
        """Checks if an active command input is queued. An active command
        is a depressed button or tilted stick.

        :return: True (on an active button) or False (no active buttons)
        :rtype: bool
        """
        if (self.current_macro_commands is not None):
            if len(self.current_macro_commands) < 2:
                return False
            else:
                return True
        elif dumps(self.controller_input) != dumps(DIRECT_INPUT_IDLE_PACKET):
            return True
        else:
            return False

    def set_protocol_input(self, state=None):

        # Act on direct input if we're not getting idle packets
        if dumps(self.controller_input) != dumps(DIRECT_INPUT_IDLE_PACKET):
            self.parse_controller_input(self.controller_input)
            self.controller_input = None

        elif (self.macro_buffer or self.current_macro or
              self.current_macro_commands):
            # Check if we can start on a new macro.
            if not self.current_macro and self.macro_buffer:
                # Preprocess command lines of current macro
                macro = self.macro_buffer.pop(0)
                self.current_macro = self.parse_macro(macro[0])
                self.current_macro_id = macro[1]

            # Check if we can load the next set of commands
            if not self.current_macro_commands and self.current_macro:
                self.current_macro_commands = (
                    self.current_macro.pop(0).strip(" ").split(" "))

                # Timing metadata extraction
                timer_length = self.current_macro_commands[-1]
                timer_length = timer_length[0:len(timer_length)-1]
                self.macro_timer_length = float(timer_length)
                self.macro_timer_start = perf_counter()

            self.set_macro_input(self.current_macro_commands)

            # Check if we're done inputting the current command
            time_delta = perf_counter() - self.macro_timer_start
            if time_delta > self.macro_timer_length:
                self.current_macro_commands = None
                # Check if we're done the current macro
                if not self.current_macro and state:
                    finished = state["finished_macros"]
                    finished.append(self.current_macro_id)
                    state["finished_macros"] = finished

    def parse_controller_input(self, controller_input):

        # Check for input validity
        if type(controller_input) != dict:
            return

        # Check if the Grip/Order menu would be closed
        if not self.exited_grip_order_menu and (
                controller_input["A"] or controller_input["B"] or controller_input["HOME"]):
            self.exited_grip_order_menu = True

        # Arrays representing the 3 button bytes in the
        # standard input report as binary.
        upper = ['0'] * 8
        shared = ['0'] * 8
        lower = ['0'] * 8
        # Upper Byte
        if controller_input["Y"]:
            upper[7] = '1'
        if controller_input["X"]:
            upper[6] = '1'
        if controller_input["B"]:
            upper[5] = '1'
        if controller_input["A"]:
            upper[4] = '1'
        if controller_input["JCL_SR"]:
            upper[3] = '1'
        if controller_input["JCL_SL"]:
            upper[2] = '1'
        if controller_input["R"]:
            upper[1] = '1'
        if controller_input["ZR"]:
            upper[0] = '1'

        # Shared byte
        if controller_input["MINUS"]:
            shared[7] = '1'
        if controller_input["PLUS"]:
            shared[6] = '1'
        if controller_input["R_STICK"]["PRESSED"]:
            shared[5] = '1'
        if controller_input["L_STICK"]["PRESSED"]:
            shared[4] = '1'
        if controller_input["HOME"]:
            shared[3] = '1'
        if controller_input["CAPTURE"]:
            shared[2] = '1'

        # Lower byte
        if controller_input["DPAD_DOWN"]:
            lower[7] = '1'
        if controller_input["DPAD_UP"]:
            lower[6] = '1'
        if controller_input["DPAD_RIGHT"]:
            lower[5] = '1'
        if controller_input["DPAD_LEFT"]:
            lower[4] = '1'
        if controller_input["JCR_SR"]:
            lower[3] = '1'
        if controller_input["JCR_SL"]:
            lower[2] = '1'
        if controller_input["L"]:
            lower[1] = '1'
        if controller_input["ZL"]:
            lower[0] = '1'

        # Analog Stick Positions
        stick_left = self.stick_ratio_to_calibrated_position(
            controller_input["L_STICK"]["X_VALUE"] / 100,
            controller_input["L_STICK"]["Y_VALUE"] / 100,
            "L_STICK"
        )
        stick_right = self.stick_ratio_to_calibrated_position(
            controller_input["R_STICK"]["X_VALUE"] / 100,
            controller_input["R_STICK"]["Y_VALUE"] / 100,
            "R_STICK"
        )

        # Converting binary strings to ints
        upper_byte = int("".join(upper), 2)
        shared_byte = int("".join(shared), 2)
        lower_byte = int("".join(lower), 2)

        self.protocol.set_button_inputs(upper_byte, shared_byte, lower_byte)
        self.protocol.set_left_stick_inputs(stick_left)
        self.protocol.set_right_stick_inputs(stick_right)

        return controller_input

    def parse_macro(self, macro):

        parsed = macro.split("\n")
        parsed = list(filter(lambda s: not s.strip() == "", parsed))
        parsed = self.parse_loops(parsed)

        return parsed

    def parse_loops(self, macro):
        parsed = []
        i = 0
        while i < len(macro):
            line = macro[i]
            if line.startswith("LOOP"):
                loop_count = int(line.split(" ")[1])
                loop_buffer = []

                # Detect delimiter and record
                if macro[i+1].startswith("\t"):
                    loop_delimiter = "\t"
                elif macro[i+1].startswith("    "):
                    loop_delimiter = "    "
                else:
                    loop_delimiter = "  "

                # Gather looping commands
                for j in range(i+1, len(macro)):
                    loop_line = macro[j]
                    if loop_line.startswith(loop_delimiter):
                        # Replace the first instance of the delimiter
                        loop_line = loop_line.replace(loop_delimiter, "", 1)
                        loop_buffer.append(loop_line)
                    # Set the new position if we either encounter the end
                    # of the loop or we reach the end of the macro
                    else:
                        i = j - 1
                        break
                    if j+1 >= len(macro):
                        i = j

                # Recursively gather other loops if present
                if any(s.startswith("LOOP") for s in loop_buffer):
                    loop_buffer = self.parse_loops(loop_buffer)
                # Multiply out the loop and concatenate
                parsed = parsed + (loop_buffer * loop_count)
            else:
                parsed.append(line)
            i += 1

        return parsed

    def set_macro_input(self, macro_input):

        # Checking if this is a wait macro command
        if len(macro_input) < 2:
            return

        # Check if the Grip/Order menu would be closed
        if not self.exited_grip_order_menu and (
                'A' in macro_input or 'B' in macro_input or 'HOME' in macro_input):
            self.exited_grip_order_menu = True

        # Arrays representing the 3 button bytes in the
        # standard input report as binary.
        upper = ['0'] * 8
        shared = ['0'] * 8
        lower = ['0'] * 8
        # Analog stick byte placeholders
        stick_left = None
        stick_right = None
        for i in range(0, len(macro_input)-1):
            button = macro_input[i]
            # Upper Byte
            if button == "Y":
                upper[7] = '1'
            elif button == "X":
                upper[6] = '1'
            elif button == "B":
                upper[5] = '1'
            elif button == "A":
                upper[4] = '1'
            elif button == "JCL_SR":
                upper[3] = '1'
            elif button == "JCL_SL":
                upper[2] = '1'
            elif button == "R":
                upper[1] = '1'
            elif button == "ZR":
                upper[0] = '1'

            # Shared byte
            elif button == "MINUS":
                shared[7] = '1'
            elif button == "PLUS":
                shared[6] = '1'
            elif button == "R_STICK_PRESS":
                shared[5] = '1'
            elif button == "L_STICK_PRESS":
                shared[4] = '1'
            elif button == "HOME":
                shared[3] = '1'
            elif button == "CAPTURE":
                shared[2] = '1'

            # Lower byte
            elif button == "DPAD_DOWN":
                lower[7] = '1'
            elif button == "DPAD_UP":
                lower[6] = '1'
            elif button == "DPAD_RIGHT":
                lower[5] = '1'
            elif button == "DPAD_LEFT":
                lower[4] = '1'
            elif button == "JCR_SR":
                lower[3] = '1'
            elif button == "JCR_SL":
                lower[2] = '1'
            elif button == "L":
                lower[1] = '1'
            elif button == "ZL":
                lower[0] = '1'

            # Analog Stick Positions
            elif button.startswith("L_STICK@"):
                stick_left = self.parse_macro_stick_position(button)
            elif button.startswith("R_STICK@"):
                stick_right = self.parse_macro_stick_position(button)

        # Converting binary strings to ints
        upper_byte = int("".join(upper), 2)
        shared_byte = int("".join(shared), 2)
        lower_byte = int("".join(lower), 2)

        self.protocol.set_button_inputs(upper_byte, shared_byte, lower_byte)
        if stick_left:
            self.protocol.set_left_stick_inputs(stick_left)
        if stick_right:
            self.protocol.set_right_stick_inputs(stick_right)

    def parse_macro_stick_position(self, stick_pos):

        stick_type = stick_pos.split("@")[0]
        positions = stick_pos.split("@")[1]
        if len(positions) < 8:
            return None

        # Converting macro to proper ratios
        sign_x = positions[0]
        ratio_x = int(positions[1:4]) / 100
        if sign_x == "-":
            ratio_x = ratio_x * -1

        sign_y = positions[4]
        ratio_y = int(positions[5:8]) / 100
        if sign_y == "-":
            ratio_y = ratio_y * -1

        calibrated_position = self.stick_ratio_to_calibrated_position(
            ratio_x, ratio_y, stick_type)

        return calibrated_position

    def stick_ratio_to_calibrated_position(self, ratio_x, ratio_y, stick_type):

        # Using the appropriate calibration values for the stick type
        if stick_type == "L_STICK":
            cal = self.LEFT_STICK_CALIBRATION
        else:
            cal = self.RIGHT_STICK_CALIBRATION

        # Converting ratios to uint16 values
        if ratio_x < 0:
            data_x_converted = (
                abs(ratio_x) * cal["min_x"] + cal["center_x"])
        else:
            data_x_converted = (
                abs(ratio_x) * cal["max_x"] + cal["center_x"])
        data_x_converted = int(round(data_x_converted))

        if ratio_y < 0:
            data_y_converted = (
                abs(ratio_y) * cal["min_y"] + cal["center_y"])
        else:
            data_y_converted = (
                abs(ratio_y) * cal["max_y"] + cal["center_y"])
        data_y_converted = int(round(data_y_converted))

        # Converting the two X/Y uint16 values to 3 uint8 Little Endian values
        # using bitshifting techniques
        converted_values = [
            # Get the last two hex digits
            data_x_converted & 0xFF,
            # Combine the last digit of the Y uint16 and the first digit
            # of the X uint16
            ((data_y_converted & 0xF) << 4) + (data_x_converted >> 8),
            # Get the first two digits of the Y uint16
            data_y_converted >> 4]

        return converted_values

    def reassign_protocol(self, protocol):

        self.protocol = protocol



from typing import Union

from .nxbt.nxbt import Buttons as NXBT_Buttons, Nxbt, PRO_CONTROLLER, find_devices_by_alias
from .wrapper import Button, NXWrapper


def check_bluetooth_address(address: str) -> None:
    """Check the validity of a given Bluetooth MAC address

    :param address: A Bluetooth MAC address
    :type address: str
    :raises ValueError: If the Bluetooth address is invalid
    """

    address_bytes = len(address.split(":"))
    if address_bytes != 6:
        raise ValueError("Invalid Bluetooth address")


def get_reconnect_target(reconnect: bool, address: str):
    if reconnect:
        reconnect_target = find_devices_by_alias("Nintendo Switch")
    elif address:
        check_bluetooth_address(address)
        reconnect_target = address
    else:
        reconnect_target = None

    return reconnect_target


class NXBTControl(NXWrapper):
    button_map = {
        Button.A:             NXBT_Buttons.A,
        Button.B:             NXBT_Buttons.B,
        Button.X:             NXBT_Buttons.X,
        Button.Y:             NXBT_Buttons.Y,
        Button.DPAD_UP:       NXBT_Buttons.DPAD_UP,
        Button.DPAD_DOWN:     NXBT_Buttons.DPAD_DOWN,
        Button.DPAD_LEFT:     NXBT_Buttons.DPAD_LEFT,
        Button.DPAD_RIGHT:    NXBT_Buttons.DPAD_RIGHT,
        Button.SHOULDER_L:    NXBT_Buttons.L,
        Button.SHOULDER_R:    NXBT_Buttons.R,
        Button.SHOULDER_ZL:   NXBT_Buttons.ZL,
        Button.SHOULDER_ZR:   NXBT_Buttons.ZR,
        Button.L_STICK_PRESS: NXBT_Buttons.L_STICK_PRESS,
        Button.R_STICK_PRESS: NXBT_Buttons.R_STICK_PRESS,
        Button.HOME:          NXBT_Buttons.HOME,
        Button.CAPTURE:       NXBT_Buttons.CAPTURE,
        Button.MINUS:         NXBT_Buttons.MINUS,
        Button.PLUS:          NXBT_Buttons.PLUS,
        Button.JCL_SR:        NXBT_Buttons.JCL_SR,
        Button.JCL_SL:        NXBT_Buttons.JCL_SL,
        Button.JCR_SR:        NXBT_Buttons.JCR_SR,
        Button.JCR_SL:        NXBT_Buttons.JCR_SL
        }

    def __init__(self, press_duration_ms: int = 50, delay_ms: int=50, reconnect: bool = False, address: str = None):
        reconnect_target = get_reconnect_target(reconnect, address)
        super().__init__(press_duration_ms=press_duration_ms, delay_ms=delay_ms)
        self.nx = Nxbt()

        # Get a list of all available Bluetooth adapters
        adapters = self.nx.get_available_adapters()
        # Prepare a list to store the indexes of the
        # created controllers.
        controller_idxs = []
        # Loop over all Bluetooth adapters and create
        # Switch Pro Controllers
        for i in range(0, len(adapters)):
            index = self.nx.create_controller(
                    PRO_CONTROLLER,
                    adapter_path=adapters[i],
                    reconnect_address=reconnect_target
                    )
            controller_idxs.append(index)

        # Select the last controller for input
        self.controller_idx = controller_idxs[-1]

    def connect(self):
        self.nx.wait_for_connection(self.controller_idx)

    def button_hold(self, button: Button, duration_ms: int):
        self.nx.press_buttons(controller_index=self.controller_idx,
                              buttons=[self.button_map[button]],
                              down=duration_ms / 1000,
                              up=self.delay_ms,
                              block=True
                              )

    def series_press(self, button_names: list[Union[Button, tuple[Button, int]]]):
        button_list = []
        for button in button_names:
            if isinstance(button, tuple):
                button_press = "{} {}s\n {}s".format(self.button_map[button[0]],
                                                     button[1] / 1000,
                                                     self.delay_ms / 1000
                                                     )
            else:
                button_press = "{} {}s\n {}s".format(self.button_map[button],
                                                     self.press_duration_ms / 1000,
                                                     self.delay_ms / 1000
                                                     )
            button_list.append(button_press)
        macro = "\n".join(button_list)
        print(macro)
        self.nx.macro(self.controller_idx, macro, block=True)

    def disconnect(self):
        pass



import socket
import fcntl
import os
import time
import queue
import logging
import traceback
import atexit
from threading import Thread
import statistics as stat

from .controller import Controller, ControllerTypes
from ..bluez import BlueZ, find_devices_by_alias
from .protocol import ControllerProtocol
from .input import InputParser
from .utils import format_msg_controller, format_msg_switch


class ControllerServer():

    def __init__(self, controller_type, adapter_path="/org/bluez/hci0",
                 state=None, task_queue=None, lock=None, colour_body=None,
                 colour_buttons=None):

        self.logger = logging.getLogger('nxbt')
        # Cache logging level to increase performance on checks
        self.logger_level = self.logger.level

        atexit.register(self._on_exit)

        if state:
            self.state = state
        else:
            self.state = {
                "state": "",
                "finished_macros": [],
                "errors": None,
                "direct_input": None
            }

        self.task_queue = task_queue

        self.controller_type = controller_type
        self.colour_body = colour_body
        self.colour_buttons = colour_buttons

        if lock:
            self.lock = lock

        self.reconnect_counter = 0

        # Intializing Bluetooth
        self.bt = BlueZ(adapter_path=adapter_path)

        self.controller = Controller(self.bt, self.controller_type)
        self.protocol = ControllerProtocol(
            self.controller_type,
            self.bt.address,
            colour_body=self.colour_body,
            colour_buttons=self.colour_buttons)

        self.input = InputParser(self.protocol)

        # Debug timekeeping storage array
        self.times = []

        # Initial reconnection overload protection
        self.tick = 1
        self.cached_msg = ''

    def run(self, reconnect_address=None):
        """Runs the mainloop of the controller server.

        :param reconnect_address: The Bluetooth MAC address of a
        previously connected to Nintendo Switch, defaults to None
        :type reconnect_address: string or list, optional
        """

        self.state["state"] = "initializing"

        try:
            # If we have a lock, prevent other controllers
            # from initializing at the same time and saturating the DBus,
            # potentially causing a kernel panic.
            if self.lock:
                self.lock.acquire()
            try:
                self.controller.setup()

                if reconnect_address:
                    try:
                        itr, ctrl = self.reconnect(reconnect_address)
                    except OSError:
                        itr, ctrl = self.connect()
                else:
                    itr, ctrl = self.connect()
            finally:
                if self.lock:
                    self.lock.release()

            self.switch_address = itr.getpeername()[0]
            self.state["last_connection"] = self.switch_address

            self.state["state"] = "connected"

            self.mainloop(itr, ctrl)

        except KeyboardInterrupt:
            pass
        except Exception:
            try:
                self.state["state"] = "crashed"
                self.state["errors"] = traceback.format_exc()
                return self.state
            except Exception as e:
                self.logger.debug("Error during graceful shutdown:")
                self.logger.debug(traceback.format_exc())

    def mainloop(self, itr, ctrl):

        duration_start = time.perf_counter()
        while True:
            # Start timing command processing
            timer_start = time.perf_counter()

            # Attempt to get output from Switch
            try:
                reply = itr.recv(50)
                if len(reply) > 40:
                    self.logger.debug(format_msg_switch(reply))
            except BlockingIOError:
                reply = None

            # Getting any inputs from the task queue
            if self.task_queue:
                try:
                    while True:
                        msg = self.task_queue.get_nowait()
                        if msg and msg["type"] == "macro":
                            self.input.buffer_macro(
                                msg["macro"], msg["macro_id"])
                        elif msg and msg["type"] == "stop":
                            self.input.stop_macro(
                                msg["macro_id"], state=self.state)
                        elif msg and msg["type"] == "clear":
                            self.input.clear_macros()
                except queue.Empty:
                    pass

            # Set Direct Input
            if self.state["direct_input"]:
                self.input.set_controller_input(self.state["direct_input"])

            self.protocol.process_commands(reply)
            self.input.set_protocol_input(state=self.state)

            msg = self.protocol.get_report()

            if self.logger_level <= logging.DEBUG and reply and len(reply) > 45:
                self.logger.debug(format_msg_controller(msg))

            try:
                # Cache the last packet to prevent overloading the switch
                # with packets on the "Change Grip/Order" menu.
                if msg[3:] != self.cached_msg:
                    itr.sendall(msg)
                    self.cached_msg = msg[3:]
                # Send a blank packet every so often to keep the Switch
                # from disconnecting from the controller.
                elif self.tick >= 132:
                    itr.sendall(msg)
                    self.tick = 0
            except BlockingIOError:
                continue
            except OSError as e:
                # Attempt to reconnect to the Switch
                itr, ctrl = self.save_connection(e)

            # Figure out how long it took to process commands
            duration_end = time.perf_counter()
            duration_elapsed = duration_end - duration_start
            duration_start = duration_end
            
            sleep_time = 1/132 - duration_elapsed
            if sleep_time >= 0:
                time.sleep(sleep_time)
            self.tick += 1

            if self.logger_level <= logging.DEBUG:
                self.times.append(duration_elapsed)
                if len(self.times) > 100:
                    self.times.pop()
                mean_time = stat.mean(self.times)

                self.logger.debug(
                    f"Tick: {self.tick}, Mean Time: {str(1/mean_time)}")


    def save_connection(self, error, state=None):

        while self.reconnect_counter < 2:
            try:
                self.logger.debug("Attempting to reconnect")
                # Reinitialize the protocol
                self.protocol = ControllerProtocol(
                    self.controller_type,
                    self.bt.address,
                    colour_body=self.colour_body,
                    colour_buttons=self.colour_buttons)
                self.input.reassign_protocol(self.protocol)
                if self.lock:
                    self.lock.acquire()
                try:
                    itr, ctrl = self.reconnect(self.switch_address)

                    received_first_message = False
                    while True:
                        # Attempt to get output from Switch
                        try:
                            reply = itr.recv(50)
                            if self.logger_level <= logging.DEBUG and len(reply) > 40:
                                self.logger.debug(format_msg_switch(reply))
                        except BlockingIOError:
                            reply = None

                        if reply:
                            received_first_message = True

                        self.protocol.process_commands(reply)
                        msg = self.protocol.get_report()

                        if self.logger_level <= logging.DEBUG and reply:
                            self.logger.debug(format_msg_controller(msg))

                        try:
                            itr.sendall(msg)
                        except BlockingIOError:
                            continue

                        # Exit pairing loop when player lights have been set and
                        # vibration has been enabled
                        if (reply and len(reply) > 45 and
                                self.protocol.vibration_enabled and self.protocol.player_number):
                            break

                        # Switch responds to packets slower during pairing
                        # Pairing cycle responds optimally on a 15Hz loop
                        if not received_first_message:
                            time.sleep(1)
                        else:
                            time.sleep(1/15)

                    self.state["state"] = "connected"
                    return itr, ctrl
                finally:
                    if self.lock:
                        self.lock.release()
            except OSError:
                self.reconnect_counter += 1
                self.logger.debug(error)
                time.sleep(0.5)

        # If we can't reconnect, transition to attempting
        # to connect to any Switch.
        self.logger.debug("Connecting to any Switch")
        self.reconnect_counter = 0

        # Reinitialize initial communication overload protections
        self.tick = 1

        # Reinitialize the protocol
        self.protocol = ControllerProtocol(
            self.controller_type,
            self.bt.address,
            colour_body=self.colour_body,
            colour_buttons=self.colour_buttons)
        self.input.reassign_protocol(self.protocol)

        # Since we were forced to attempt a reconnection
        # we need to press the L/SL and R/SR buttons before
        # we can proceed with any input.
        if self.controller_type == ControllerTypes.PRO_CONTROLLER:
            self.input.current_macro_commands = "L R 0.0s".strip(" ").split(" ")
        elif self.controller_type == ControllerTypes.JOYCON_L:
            self.input.current_macro_commands = "JCL_SL JCL_SR 0.0s".strip(" ").split(" ")
        elif self.controller_type == ControllerTypes.JOYCON_R:
            self.input.current_macro_commands = "JCR_SL JCR_SR 0.0s".strip(" ").split(" ")

        if self.lock:
            self.lock.acquire()
        try:
            itr, ctrl = self.connect()
        finally:
            if self.lock:
                self.lock.release()

        self.state["state"] = "connected"

        self.switch_address = itr.getsockname()[0]

        return itr, ctrl

    def connection_reset_watchdog(self):

        connected_devices = []
        connected_devices_count = {}
        while self._crw_running:
            paths = self.bt.find_connected_devices(alias_filter="Nintendo Switch")
            # Keep track of Switches that connect
            if len(paths) > 0:
                connected_devices = list(set(connected_devices + paths))
            
            # Increment a counter if a Switch connected and disconnected
            disconnected = list(set(connected_devices) - set(paths))
            if len(disconnected) > 0:
                for path in disconnected:
                    if path not in connected_devices_count.keys():
                        connected_devices_count[path] = 1
                    else:
                        connected_devices_count[path] += 1
                connected_devices = list(set(connected_devices) - set(disconnected))
            
            # Delete Switches that connect/disconnect twice.
            # This behaviour is characteristic of connection issues and is corrected
            # by removing the Switch's connection to the system.
            if len(connected_devices_count.keys()) > 0:
                for key in connected_devices_count.keys():
                    if connected_devices_count[key] >= 2:
                        self.logger.debug(
                            "A Nintendo Switch disconnected. Resetting Connection...")
                        self.logger.debug(f"Removing {str(key)}")
                        self.bt.remove_device(key)
                        connected_devices_count[key] = 0

            time.sleep(0.1)

    def connect(self):
        """Configures as a specified controller, pairs with a Nintendo Switch,
        and creates/accepts sockets for communication with the Switch.
        """

        # The controller server will continue attempting to connect
        # to any Nintendo Switch until the connection procedure fully
        # succeeds. This prevents situations where the Switch will
        # disconnect during a connection.
        while True:
            try:
                self.state["state"] = "connecting"

                # Creating control and interrupt sockets
                s_ctrl = socket.socket(
                    family=socket.AF_BLUETOOTH,
                    type=socket.SOCK_SEQPACKET,
                    proto=socket.BTPROTO_L2CAP)
                s_itr = socket.socket(
                    family=socket.AF_BLUETOOTH,
                    type=socket.SOCK_SEQPACKET,
                    proto=socket.BTPROTO_L2CAP)

                # Setting up HID interrupt/control sockets
                try:
                    s_ctrl.bind((self.bt.address, 17))
                    s_itr.bind((self.bt.address, 19))
                except OSError:
                    s_ctrl.bind((socket.BDADDR_ANY, 17))
                    s_itr.bind((socket.BDADDR_ANY, 19))

                s_itr.listen(1)
                s_ctrl.listen(1)

                self.bt.set_discoverable(True)

                # WARNING:
                # A device's class must be set **AFTER** discoverability
                # is set. If it is set before or in a similar timeframe,
                # the class will be reset to the default value.
                self.bt.set_class("0x02508")

                self._crw_running = True
                crw = Thread(target = self.connection_reset_watchdog)
                crw.start()

                itr, itr_address = s_itr.accept()
                ctrl, ctrl_address = s_ctrl.accept()

                self._crw_running = False

                # Send an empty input report to the Switch to prompt a reply
                self.protocol.process_commands(None)
                msg = self.protocol.get_report()
                itr.sendall(msg)

                # Setting interrupt connection as non-blocking.
                # In this case, non-blocking means it throws a "BlockingIOError"
                # for sending and receiving, instead of blocking.
                fcntl.fcntl(itr, fcntl.F_SETFL, os.O_NONBLOCK)

                # Mainloop
                received_first_message = False
                while True:
                    # Attempt to get output from Switch
                    try:
                        reply = itr.recv(50)
                        if self.logger_level <= logging.DEBUG and len(reply) > 40:
                            self.logger.debug(format_msg_switch(reply))
                    except BlockingIOError:
                        reply = None

                    if reply:
                        received_first_message = True

                    self.protocol.process_commands(reply)
                    msg = self.protocol.get_report()

                    if self.logger_level <= logging.DEBUG and reply:
                        self.logger.debug(format_msg_controller(msg))

                    try:
                        itr.sendall(msg)
                    except BlockingIOError:
                        continue

                    # Exit pairing loop when player lights have been set and
                    # vibration has been enabled
                    if (reply and len(reply) > 45 and
                            self.protocol.vibration_enabled and self.protocol.player_number):
                        break

                    # Switch responds to packets slower during pairing
                    # Pairing cycle responds optimally on a 15Hz loop
                    if not received_first_message:
                        time.sleep(1)
                    else:
                        time.sleep(1/15)
                
                break
            except OSError as e:
                self.logger.debug(e)

        self.input.exited_grip_order_menu = False

        return itr, ctrl

    def reconnect(self, reconnect_address):
        """Attempts to reconnect with a Switch at the given address.

        :param reconnect_address: The Bluetooth MAC address of the Switch
        :type reconnect_address: string or list
        """

        def recreate_sockets():
            # Creating control and interrupt sockets
            ctrl = socket.socket(
                family=socket.AF_BLUETOOTH,
                type=socket.SOCK_SEQPACKET,
                proto=socket.BTPROTO_L2CAP)
            itr = socket.socket(
                family=socket.AF_BLUETOOTH,
                type=socket.SOCK_SEQPACKET,
                proto=socket.BTPROTO_L2CAP)

            return itr, ctrl

        self.state["state"] = "reconnecting"

        itr = None
        ctrl = None
        if type(reconnect_address) == list:
            for address in reconnect_address:
                test_itr, test_ctrl = recreate_sockets()
                try:
                    # Setting up HID interrupt/control sockets
                    test_ctrl.connect((address, 17))
                    test_itr.connect((address, 19))

                    itr = test_itr
                    ctrl = test_ctrl
                except OSError:
                    test_itr.close()
                    test_ctrl.close()
                    pass
        elif type(reconnect_address) == str:
            test_itr, test_ctrl = recreate_sockets()

            # Setting up HID interrupt/control sockets
            test_ctrl.connect((reconnect_address, 17))
            test_itr.connect((reconnect_address, 19))

            itr = test_itr
            ctrl = test_ctrl

        if not itr and not ctrl:
            raise OSError("Unable to reconnect to sockets at the given address(es)",
                          reconnect_address)

        fcntl.fcntl(itr, fcntl.F_SETFL, os.O_NONBLOCK)

        # Send an empty input report to the Switch to prompt a reply
        self.protocol.process_commands(None)
        msg = self.protocol.get_report()
        itr.sendall(msg)

        # Setting interrupt connection as non-blocking
        # In this case, non-blocking means it throws a "BlockingIOError"
        # for sending and receiving, instead of blocking
        fcntl.fcntl(itr, fcntl.F_SETFL, os.O_NONBLOCK)

        return itr, ctrl

    def _on_exit(self):
        self.bt.reset_address()



import asyncio
import logging
import socket

import dbus
import pkg_resources

from joycontrol import utils
from joycontrol.device import HidDevice
from joycontrol.report import InputReport
from joycontrol.transport import L2CAP_Transport

PROFILE_PATH = pkg_resources.resource_filename('joycontrol', 'profile/sdp_record_hid.xml')
logger = logging.getLogger(__name__)


async def _send_empty_input_reports(transport):
    report = InputReport()
    for i in range(10):
        await transport.write(report)
        await asyncio.sleep(1)


async def create_hid_server(protocol_factory, ctl_psm=17, itr_psm=19, device_id=None, reconnect_bt_addr=None,
                            capture_file=None):
    """
    :param protocol_factory: Factory function returning a ControllerProtocol instance
    :param ctl_psm: hid control channel port
    :param itr_psm: hid interrupt channel port
    :param device_id: ID of the bluetooth adapter.
                      Integer matching the digit in the hci* notation (e.g. hci0, hci1, ...) or
                      Bluetooth mac address in string notation of the adapter (e.g. "FF:FF:FF:FF:FF:FF").
                      If None, choose any device.
                      Note: Selection of adapters may currently not work if the bluez "input" plugin is enabled.
    :param reconnect_bt_addr: The Bluetooth address of the console that was previously connected. Defaults to None.
                      If None, a new hid server will be started for the initial paring.
                      Otherwise, the function assumes an initial pairing with the console was already done
                      and reconnects to the provided Bluetooth address.
    :param capture_file: opened file to log incoming and outgoing messages
    :returns transport for input reports and protocol which handles incoming output reports
    """
    protocol = protocol_factory()

    if reconnect_bt_addr is None:
        ctl_sock = socket.socket(socket.AF_BLUETOOTH, socket.SOCK_SEQPACKET, socket.BTPROTO_L2CAP)
        itr_sock = socket.socket(socket.AF_BLUETOOTH, socket.SOCK_SEQPACKET, socket.BTPROTO_L2CAP)
        ctl_sock.setblocking(False)
        itr_sock.setblocking(False)
        ctl_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        itr_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        
        try:
            hid = HidDevice(device_id=device_id)

            ctl_sock.bind((hid.address, ctl_psm))
            itr_sock.bind((hid.address, itr_psm))
        except OSError as err:
            logger.warning(err)
            # If the ports are already taken, this probably means that the bluez "input" plugin is enabled.
            logger.warning('Fallback: Restarting bluetooth due to incompatibilities with the bluez "input" plugin. '
                           'Disable the plugin to avoid issues. See https://github.com/mart1nro/joycontrol/issues/8.')
            # HACK: To circumvent incompatibilities with the bluetooth "input" plugin, we need to restart Bluetooth here.
            # The Switch does not connect to the sockets if we don't.
            # For more info see: https://github.com/mart1nro/joycontrol/issues/8
            logger.info('Restarting bluetooth service...')
            await utils.run_system_command('systemctl restart bluetooth.service')
            await asyncio.sleep(1)

            hid = HidDevice(device_id=device_id)

            ctl_sock.bind((socket.BDADDR_ANY, ctl_psm))
            itr_sock.bind((socket.BDADDR_ANY, itr_psm))

        ctl_sock.listen(1)
        itr_sock.listen(1)

        hid.powered(True)
        hid.pairable(True)
        
        # setting bluetooth adapter name to the device we wish to emulate
        await hid.set_name(protocol.controller.device_name())

        logger.info('Advertising the Bluetooth SDP record...')
        try:
            HidDevice.register_sdp_record(PROFILE_PATH)
        except dbus.exceptions.DBusException as dbus_err:
            # Already registered (If multiple controllers are being emulated and this method is called consecutive times)
            logger.debug(dbus_err)

        # set the device class to "Gamepad/joystick"
        await hid.set_class()

        # start advertising
        hid.discoverable()

        logger.info('Waiting for Switch to connect... Please open the "Change Grip/Order" menu.')

        loop = asyncio.get_event_loop()
        client_ctl, ctl_address = await loop.sock_accept(ctl_sock)
        logger.info(f'Accepted connection at psm {ctl_psm} from {ctl_address}')
        client_itr, itr_address = await loop.sock_accept(itr_sock)
        logger.info(f'Accepted connection at psm {itr_psm} from {itr_address}')
        assert ctl_address[0] == itr_address[0]

        # stop advertising
        hid.discoverable(False)
        hid.pairable(False)

    else:
        # Reconnection to reconnect_bt_addr
        client_ctl = socket.socket(socket.AF_BLUETOOTH, socket.SOCK_SEQPACKET, socket.BTPROTO_L2CAP)
        client_itr = socket.socket(socket.AF_BLUETOOTH, socket.SOCK_SEQPACKET, socket.BTPROTO_L2CAP)
        client_ctl.connect((reconnect_bt_addr, ctl_psm))
        client_itr.connect((reconnect_bt_addr, itr_psm))
        client_ctl.setblocking(False)
        client_itr.setblocking(False)

    # create transport for the established connection and activate the HID protocol
    transport = L2CAP_Transport(asyncio.get_event_loop(), protocol, client_itr, client_ctl, 50, capture_file=capture_file)
    protocol.connection_made(transport)

    # HACK: send some empty input reports until the Switch decides to reply
    future = asyncio.ensure_future(_send_empty_input_reports(transport))
    await protocol.wait_for_output_report()
    """
    future.cancel()
    try:
        await future
    except asyncio.CancelledError:
        pass
    """

    return protocol.transport, protocol



import subprocess
import re
import os
import time
import logging
from shutil import which
import random
from pathlib import Path

import dbus


SERVICE_NAME = "org.bluez"
BLUEZ_OBJECT_PATH = "/org/bluez"
ADAPTER_INTERFACE = SERVICE_NAME + ".Adapter1"
PROFILEMANAGER_INTERFACE = SERVICE_NAME + ".ProfileManager1"
DEVICE_INTERFACE = SERVICE_NAME + ".Device1"


def find_object_path(bus, service_name, interface_name, object_name=None):
    """Searches for a D-Bus object path that contains a specified interface
    under a specified service.

    :param bus: A DBus object used to access the DBus.
    :type bus: DBus
    :param service_name: The name of a D-Bus service to search for the
    object path under.
    :type service_name: string
    :param interface_name: The name of a D-Bus interface to search for
    within objects under the specified service.
    :type interface_name: string
    :param object_name: The name or ending of the object path,
    defaults to None
    :type object_name: string, optional
    :return: The D-Bus object path or None, if no matching object
    can be found
    :rtype: string
    """

    manager = dbus.Interface(
        bus.get_object(service_name, "/"),
        "org.freedesktop.DBus.ObjectManager")

    # Iterating over objects under the specified service
    # and searching for the specified interface
    for path, ifaces in manager.GetManagedObjects().items():
        managed_interface = ifaces.get(interface_name)
        if managed_interface is None:
            continue
        # If the object name wasn't specified or it matches
        # the interface address or the path ending
        elif (not object_name or
                object_name == managed_interface["Address"] or
                path.endswith(object_name)):
            obj = bus.get_object(service_name, path)
            return dbus.Interface(obj, interface_name).object_path

    return None


def find_objects(bus, service_name, interface_name):
    """Searches for D-Bus objects that contain a specified interface
    under a specified service.

    :param bus: A DBus object used to access the DBus.
    :type bus: DBus
    :param service_name: The name of a D-Bus service to search for the
    object path under.
    :type service_name: string
    :param interface_name: The name of a D-Bus interface to search for
    within objects under the specified service.
    :type interface_name: string
    :return: The D-Bus object paths matching the arguments
    :rtype: array
    """

    manager = dbus.Interface(
        bus.get_object(service_name, "/"),
        "org.freedesktop.DBus.ObjectManager")
    paths = []

    # Iterating over objects under the specified service
    # and searching for the specified interface within them
    for path, ifaces in manager.GetManagedObjects().items():
        managed_interface = ifaces.get(interface_name)
        if managed_interface is None:
            continue
        else:
            obj = bus.get_object(service_name, path)
            path = str(dbus.Interface(obj, interface_name).object_path)
            paths.append(path)

    return paths


def toggle_clean_bluez(toggle):
    """Enables or disables all BlueZ plugins,
    BlueZ compatibility mode, and removes all extraneous
    SDP Services offered.
    Requires root user to be run. The units and Bluetooth
    service will not be restarted if the input plugin
    already matches the toggle.

    :param toggle: A boolean element indicating if BlueZ 
    should be cleaned (True) or not (False)
    :type toggle: boolean
    :raises PermissionError: If the user is not root
    :raises Exception: If the units can't be reloaded
    :raises Exception: If sdptool, hciconfig, or hcitool are not available.
    """

    service_path = "/lib/systemd/system/bluetooth.service"
    override_dir = Path("/run/systemd/system/bluetooth.service.d")
    override_path = override_dir / "nxbt.conf"

    if toggle:
        if override_path.is_file():
            # Override exist, no need to restart bluetooth
            return

        with open(service_path) as f:
            for line in f:
                if line.startswith("ExecStart="):
                    exec_start = line.strip() + " --compat --noplugin=*"
                    break
            else:
                raise Exception("systemd service file doesn't have a ExecStart line")

        override = f"[Service]\nExecStart=\n{exec_start}"

        override_dir.mkdir(parents=True, exist_ok=True)
        with override_path.open("w") as f:
            f.write(override)
    else:
        try:
            os.remove(override_path)
        except FileNotFoundError:
            # Override doesn't exist, no need to restart bluetooth
            return

    # Reload units
    _run_command(["systemctl", "daemon-reload"])

    # Reload the bluetooth service with input disabled
    _run_command(["systemctl", "restart", "bluetooth"])

    # Kill a bit of time here to ensure all services have restarted
    time.sleep(0.5)


def clean_sdp_records():
    """Cleans all SDP Records from BlueZ with sdptool

    :raises Exception: On CLI error or sdptool missing
    """
    # TODO: sdptool is deprecated in BlueZ 5. This should ideally
    # use the DBus API, however, bugs seemingly exist with the
    # UnregisterProfile interface.

    # Check if sdptool is available for use
    if which("sdptool") is None:
        raise Exception("sdptool is not available on this system." +
                        "If you can, please install this tool, as " +
                        "it is required for proper functionality.")

    # Enable Read/Write to the SDP server. This is a remedy for a 
    # compatibility mode bug introduced in later versions of BlueZ 5
    _run_command(["chmod", "777", "/var/run/sdp"])

    # Identify/List all SDP services available with sdptool
    result = _run_command(['sdptool', 'browse', 'local']).stdout.decode('utf-8')
    if result is None or len(result.split('\n\n')) < 1:
        return
    
    # Record all service record handles
    exceptions = ["PnP Information"]
    service_rec_handles = []
    for rec in result.split('\n\n'):
        # Skip if exception is in record
        exception_found = False
        for exception in exceptions:
            if exception in rec:
                exception_found = True
                break
        if exception_found:
            continue

        # Read lines and add Record Handles to the list
        for line in rec.split('\n'):
            if "Service RecHandle" in line:
                service_rec_handles.append(line.split(" ")[2])
    
    # Delete all found service records
    if len(service_rec_handles) > 0:
        for record_handle in service_rec_handles:
            _run_command(['sdptool', 'del', record_handle])


def _run_command(command):
    """Runs a specified command on the shell of the system.
    If the command is run unsuccessfully, an error is raised.
    The command must be in the form of an array with each term
    individually listed. Eg: ["which", "bash"]

    :param command: A list of command terms
    :type command: list
    :raises Exception: On command failure or error
    """
    result = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)

    cmd_err = result.stderr.decode("utf-8").replace("\n", "")
    if cmd_err != "":
        raise Exception(cmd_err)
    
    return result


def get_random_controller_mac():
    """Generates a random Switch-compliant MAC address
    """
    def seg():
        random_number = random.randint(0,255)
        hex_number = str(hex(random_number))
        hex_number = hex_number[2:].upper()
        return str(hex_number)
    
    return f"7C:BB:8A:{seg()}:{seg()}:{seg()}"


def replace_mac_addresses(adapter_paths, addresses):
    """Replaces a list of adapter's Bluetooth MAC addresses
    with Switch-compliant Controller MAC addresses. If the
    addresses argument is specified, the adapter path's
    MAC addresses will be reset to respective (index-wise)
    address in the list.

    :param adapter_paths: A list of Bluetooth adapter paths
    :type adapter_paths: list
    :param addresses: A list of Bluetooth MAC addresses,
    defaults to False
    :type addresses: bool, optional
    """
    if which("hcitool") is None:
        raise Exception("hcitool is not available on this system." +
                        "If you can, please install this tool, as " +
                        "it is required for proper functionality.")
    if which("hciconfig") is None:
        raise Exception("hciconfig is not available on this system." +
                        "If you can, please install this tool, as " +
                        "it is required for proper functionality.")

    if addresses:
        assert len(addresses) == len(adapter_paths)

    for i in range(len(adapter_paths)):
        adapter_id = adapter_paths[i].split('/')[-1]
        mac = addresses[i].split(':')
        cmds = ['hcitool', '-i', adapter_id, 'cmd', '0x3f', '0x001',
                f'0x{mac[5]}',f'0x{mac[4]}',f'0x{mac[3]}',f'0x{mac[2]}',
                f'0x{mac[1]}',f'0x{mac[0]}']
        _run_command(cmds)
        _run_command(['hciconfig', adapter_id, 'reset'])


def find_devices_by_alias(alias, return_path=False, created_bus=None):
    """Finds the Bluetooth addresses of devices
    that have a specified Bluetooth alias. Aliases
    are converted to uppercase before comparison
    as BlueZ usually converts aliases to uppercase.

    :param address: The Bluetooth MAC address
    :type address: string
    :return: The path to the D-Bus object or None
    :rtype: string or None
    """

    if created_bus is not None:
        bus = created_bus
    else:
        bus = dbus.SystemBus()
    # Find all connected/paired/discovered devices
    devices = find_objects(
        bus,
        SERVICE_NAME,
        DEVICE_INTERFACE)

    addresses = []
    matching_paths = []
    for path in devices:
        # Get the device's address and paired status
        device_props = dbus.Interface(
            bus.get_object(SERVICE_NAME, path),
            "org.freedesktop.DBus.Properties")
        device_alias = device_props.Get(
            DEVICE_INTERFACE,
            "Alias").upper()
        device_addr = device_props.Get(
            DEVICE_INTERFACE,
            "Address").upper()

        # Check for an address match
        if device_alias.upper() == alias.upper():
            addresses.append(device_addr)
            matching_paths.append(path)

    # Close the dbus connection if we created one
    if created_bus is None:
        bus.close()

    if return_path:
        return addresses, matching_paths
    else:
        return addresses


def disconnect_devices_by_alias(alias, created_bus=None):
    """Disconnects all devices matching an alias.

    :param alias: The device's alias
    :type alias: string
    """

    if created_bus is not None:
        bus = created_bus
    else:
        bus = dbus.SystemBus()
    # Find all connected/paired/discovered devices
    devices = find_objects(
        bus,
        SERVICE_NAME,
        DEVICE_INTERFACE)

    addresses = []
    matching_paths = []
    for path in devices:
        # Get the device's address and paired status
        device_props = dbus.Interface(
            bus.get_object(SERVICE_NAME, path),
            "org.freedesktop.DBus.Properties")
        device_alias = device_props.Get(
            DEVICE_INTERFACE,
            "Alias").upper()

        # Check for an alias match
        if device_alias.upper() == alias.upper():
            device = dbus.Interface(
                bus.get_object(SERVICE_NAME, path),
                DEVICE_INTERFACE)
            try:
                device.Disconnect()
            except Exception as e:
                print(e)

    # Close the dbus connection if we created one
    if created_bus is None:
        bus.close()


class BlueZ():
    """Exposes the BlueZ D-Bus API as a Python object.
    """

    def __init__(self, adapter_path="/org/bluez/hci0"):

        self.logger = logging.getLogger('nxbt')

        self.bus = dbus.SystemBus()
        self.device_path = adapter_path

        # If we weren't able to find an adapter with the specified ID,
        # try to find any usable Bluetooth adapter
        if self.device_path is None:
            self.device_path = find_object_path(
                self.bus,
                SERVICE_NAME,
                ADAPTER_INTERFACE)

        # If we aren't able to find an adapter still
        if self.device_path is None:
            raise Exception("Unable to find a bluetooth adapter")

        # Load the adapter's interface
        self.logger.debug(f"Using adapter under object path: {self.device_path}")
        self.device = dbus.Interface(
            self.bus.get_object(
                SERVICE_NAME,
                self.device_path),
            "org.freedesktop.DBus.Properties")

        self.device_id = self.device_path.split("/")[-1]

        # Load the ProfileManager interface
        self.profile_manager = dbus.Interface(self.bus.get_object(
            SERVICE_NAME, BLUEZ_OBJECT_PATH),
            PROFILEMANAGER_INTERFACE)

        self.adapter = dbus.Interface(
            self.bus.get_object(
                SERVICE_NAME,
                self.device_path),
            ADAPTER_INTERFACE)

    @property
    def address(self):
        """Gets the Bluetooth MAC address of the Bluetooth adapter.

        :return: The Bluetooth Adapter's MAC address
        :rtype: string
        """

        return self.device.Get(ADAPTER_INTERFACE, "Address").upper()

    def set_address(self, mac):
        """Sets the Bluetooth MAC address of the Bluetooth adapter.
        The hciconfig CLI is required for setting the address.
        For changes to apply, the Bluetooth interface needs to be
        restarted.

        :param mac: A Bluetooth MAC address in 
        the form of "XX:XX:XX:XX:XX:XX
        :type mac: str
        :raises PermissionError: On run as non-root user
        :raises Exception: On CLI errors
        """
        if which("hcitool") is None:
            raise Exception("hcitool is not available on this system." +
                            "If you can, please install this tool, as " +
                            "it is required for proper functionality.")
        # Reverse MAC (element position-wise) for use with hcitool
        mac = mac.split(":")
        cmds = ['hcitool', '-i', self.device_id, 'cmd', '0x3f', '0x001',
                f'0x{mac[5]}',f'0x{mac[4]}',f'0x{mac[3]}',f'0x{mac[2]}',
                f'0x{mac[1]}',f'0x{mac[0]}']
        _run_command(cmds)
        _run_command(['hciconfig', self.device_id, 'reset'])

    def set_class(self, device_class):
        if which("hciconfig") is None:
            raise Exception("hciconfig is not available on this system." +
                            "If you can, please install this tool, as " +
                            "it is required for proper functionality.")
        _run_command(['hciconfig', self.device_id, 'class', device_class])

    def reset_adapter(self):
        if which("hciconfig") is None:
            raise Exception("hciconfig is not available on this system." +
                            "If you can, please install this tool, as " +
                            "it is required for proper functionality.")
        _run_command(['hciconfig', self.device_id, 'reset'])

    @property
    def name(self):
        """Gets the name of the Bluetooth adapter.

        :return: The name of the Bluetooth adapter.
        :rtype: string
        """

        return self.device.Get(ADAPTER_INTERFACE, "Name")

    @property
    def alias(self):
        """Gets the alias of the Bluetooth adapter. This value is used
        as the "friendly" name of the adapter when communicating over
        Bluetooth.

        :return: The adapter's alias
        :rtype: string
        """

        return self.device.Get(ADAPTER_INTERFACE, "Alias")

    def set_alias(self, value):
        """Asynchronously sets the alias of the Bluetooth adapter.
        If you wish to check the set value, a time delay is needed
        before the alias getter is run.

        :param value: The new value to be set as the adapter's alias
        :type value: string
        """

        self.device.Set(ADAPTER_INTERFACE, "Alias", value)

    @property
    def pairable(self):
        """Gets the pairable status of the Bluetooth adapter.

        :return: A boolean value representing if the adapter is set as
        pairable or not
        :rtype: boolean
        """

        return bool(self.device.Get(ADAPTER_INTERFACE, "Pairable"))

    def set_pairable(self, value):
        """Sets the pariable boolean status of the Bluetooth adapter.

        :param value: A boolean value representing if the adapter is
        pairable or not.
        :type value: boolean
        """

        dbus_value = dbus.Boolean(value)
        self.device.Set(ADAPTER_INTERFACE, "Pairable", dbus_value)

    @property
    def pairable_timeout(self):
        """Gets the timeout time (in seconds) for how long the adapter
        should remain as pairable. Defaults to 0 (no timeout).

        :return: The pairable timeout in seconds
        :rtype: int
        """

        return self.device.Get(ADAPTER_INTERFACE, "PairableTimeout")

    def set_pairable_timeout(self, value):
        """Sets the timeout time (in seconds) for the pairable property.

        :param value: The pairable timeout value in seconds
        :type value: int
        """

        dbus_value = dbus.UInt32(value)
        self.device.Set(ADAPTER_INTERFACE, "PairableTimeout", dbus_value)

    @property
    def discoverable(self):
        """Gets the discoverable status of the Bluetooth adapter

        :return: The boolean status of the discoverable status
        :rtype: boolean
        """

        return bool(self.device.Get(ADAPTER_INTERFACE, "Discoverable"))

    def set_discoverable(self, value):
        """Sets the discoverable boolean status of the Bluetooth adapter.

        :param value: A boolean value representing if the Bluetooth adapter
        is discoverable or not.
        :type value: boolean
        """

        dbus_value = dbus.Boolean(value)
        self.device.Set(ADAPTER_INTERFACE, "Discoverable", dbus_value)

    @property
    def discoverable_timeout(self):
        """Gets the timeout time (in seconds) for how long the adapter
        should remain as discoverable. Defaults to 180 (3 minutes).

        :return: The discoverable timeout in seconds
        :rtype: int
        """

        return self.device.Get(ADAPTER_INTERFACE, "DiscoverableTimeout")

    def set_discoverable_timeout(self, value):
        """Sets the discoverable time (in seconds) for the discoverable
        property. Setting this property to 0 results in an infinite
        discoverable timeout.

        :param value: The discoverable timeout value in seconds
        :type value: int
        """

        dbus_value = dbus.UInt32(value)
        self.device.Set(
            ADAPTER_INTERFACE,
            "DiscoverableTimeout",
            dbus_value)

    @property
    def device_class(self):
        """Gets the Bluetooth class of the device. This represents what type
        of device this reporting as (Ex: Gamepad, Headphones, etc).

        :return: A 32-bit hexadecimal Integer representing the
        Bluetooth Code for a given device type.
        :rtype: string
        """

        # This is another hacky bit. We're using hciconfig here instead
        # of the D-Bus API so that results match the setter. See the
        # setter for further justification on using hciconfig.
        result = subprocess.run(
            ["hciconfig", self.device_id, "class"],
            stdout=subprocess.PIPE)
        device_class = result.stdout.decode("utf-8").split("Class: ")[1][0:8]

        return device_class

    def set_device_class(self, device_class):
        """Sets the Bluetooth class of the device. This represents what type
        of device this reporting as (Ex: Gamepad, Headphones, etc).
        Note: To work this function *MUST* be run as the super user. An
        exception is returned if this function is run without elevation.

        :param device_class: A 32-bit Hexadecimal integer
        :type device_class: string
        :raises PermissionError: If user is not root
        :raises ValueError: If the device class is not length 8
        :raises Exception: On inability to set class
        """

        if os.geteuid() != 0:
            raise PermissionError("The device class must be set as root")

        if len(device_class) != 8:
            raise ValueError("Device class must be length 8")

        # This is a bit of a hack. BlueZ allows you to set this value, however,
        # a config file needs to filled and the BT daemon restarted. This is a
        # good compromise but requires super user privileges. Not ideal.
        result = subprocess.run(
            ["hciconfig", self.device_id, "class", device_class],
            stderr=subprocess.PIPE)

        # Checking if there was a problem setting the device class
        cmd_err = result.stderr.decode("utf-8").replace("\n", "")
        if cmd_err != "":
            raise Exception(cmd_err)

    @property
    def powered(self):
        """The powered state of the adapter (on/off) as a boolean value.

        :return: A boolean representing the powered state of the adapter.
        :rtype: boolean
        """

        return bool(self.device.Get(ADAPTER_INTERFACE, "Powered"))

    def set_powered(self, value):
        """Switches the adapter on or off.

        :param value: A boolean value switching the adapter on or off
        :type value: boolean
        """

        dbus_value = dbus.Boolean(value)
        self.device.Set(ADAPTER_INTERFACE, "Powered", dbus_value)

    def register_profile(self, profile_path, uuid, opts):
        """Registers an SDP record on the BlueZ SDP server.

        Options (non-exhaustive, refer to BlueZ docs for
        the complete list):

        - Name: Human readable name of the profile

        - Role: Specifies precise local role. Either "client"
        or "servier".

        - RequireAuthentication: A boolean value indicating if
        pairing is required before connection.

        - RequireAuthorization: A boolean value indiciating if
        authorization is needed before connection.

        - AutoConnect: A boolean value indicating whether a
        connection can be forced if a client UUID is present.

        - ServiceRecord: An XML SDP record as a string.

        :param profile_path: The path for the SDP record
        :type profile_path: string
        :param uuid: The UUID for the SDP record
        :type uuid: string
        :param opts: The options for the SDP server
        :type opts: dict
        """

        return self.profile_manager.RegisterProfile(profile_path, uuid, opts)

    def unregister_profile(self, profile):
        """Unregisters a given SDP record from the BlueZ SDP server.

        :param profile: A SDP record profile object
        :type profile: Profile
        """

        self.profile_manager.UnregisterProfile(profile)

    def reset(self):
        """Restarts the Bluetooth Service

        :raises Exception: If the bluetooth service can't be restarted
        """

        result = subprocess.run(
            ["systemctl", "restart", "bluetooth"],
            stderr=subprocess.PIPE)

        cmd_err = result.stderr.decode("utf-8").replace("\n", "")
        if cmd_err != "":
            raise Exception(cmd_err)

        self.device = dbus.Interface(
            self.bus.get_object(
                SERVICE_NAME,
                self.device_path),
            "org.freedesktop.DBus.Properties")
        self.profile_manager = dbus.Interface(
            self.bus.get_object(
                SERVICE_NAME,
                BLUEZ_OBJECT_PATH),
            PROFILEMANAGER_INTERFACE)

    def get_discovered_devices(self):
        """Gets a dict of all discovered (or previously discovered
        and connected) devices. The key is the device's dbus object
        path and the values are the device's properties.

        The following is a non-exhaustive list of the properties a
        device dictionary can contain:
        - "Address": The Bluetooth address
        - "Alias": The friendly name of the device
        - "Paired": Whether the device is paired
        - "Connected": Whether the device is presently connected
        - "UUIDs": The services a device provides

        :return: A dictionary of all discovered devices
        :rtype: dictionary
        """

        bluez_objects = dbus.Interface(
            self.bus.get_object(SERVICE_NAME, "/"),
            "org.freedesktop.DBus.ObjectManager")

        devices = {}
        objects = bluez_objects.GetManagedObjects()
        for path, interfaces in list(objects.items()):
            if DEVICE_INTERFACE in interfaces:
                devices[str(path)] = interfaces[DEVICE_INTERFACE]

        return devices

    def discover_devices(self, alias=None, timeout=10, callback=None):
        """Runs a device discovery of the timeout length (in seconds)
        on the adapter. If specified, a callback is run, every second,
        and passed an updated list of discovered devices. An alias
        can be specified to filter discovered devices.

        The following is a non-exhaustive list of the properties a
        device dictionary can contain:
        - "Address": The Bluetooth address
        - "Alias": The friendly name of the device
        - "Paired": Whether the device is paired
        - "Connected": Whether the device is presently connected
        - "UUIDs": The services a device provides

        :param alias: The alias of a bluetooth device, defaults to None
        :type alias: string, optional
        :param timeout: The discovery timeout in seconds, defaults to 10
        :type timeout: int, optional
        :param callback: A callback function, defaults to None
        :type callback: function, optional
        :return: A dictionary of discovered devices with the object path
        as the key and the device properties as the dictionary properties
        :rtype: dictionary
        """

        # TODO: Device discovery still needs work. Currently, devices
        # are added as DBus objects while device discovery runs, however,
        # added devices linger after discovery stops. This means a device
        # can become unpairable, still show up on a new discovery session,
        # and throw an error when an attempt is made to pair it. Using DBus
        # signals ("interface added"/"property changed") does not solve
        # this issue.

        # Get all devices that have been previously discovered
        devices = self.get_discovered_devices()

        # Start discovering new devices and loop
        self.set_powered(True)
        self.set_pairable(True)
        self.adapter.StartDiscovery()
        try:
            for i in range(0, timeout):
                time.sleep(1)

                new_devices = self.get_discovered_devices()
                # Shallowly merging dictionaries. Latter dictionary
                # overrides the former. Requires Python 3.5
                devices = {**devices, **new_devices}

                if callback:
                    callback(devices)
        finally:
            self.adapter.StopDiscovery()
            time.sleep(1)

        # Filter out paired devices or devices that don't
        # match a specified alias.
        filtered_devices = {}
        for key in devices.keys():
            # Filter for devices matching alias, if specified
            if "Alias" not in devices[key].keys():
                continue
            if alias and not alias == devices[key]["Alias"]:
                continue

            # Filter for paired devices
            if "Paired" not in devices[key].keys():
                continue
            if devices[key]["Paired"]:
                continue

            filtered_devices[key] = devices[key]

        return filtered_devices

    def pair_device(self, device_path):
        """Pairs a discovered device at a given DBus object path.

        :param device_path: The D-Bus object path to the device
        :type device_path: string
        """

        device = dbus.Interface(
            self.bus.get_object(
                SERVICE_NAME,
                device_path),
            DEVICE_INTERFACE)
        device.Pair()

    def connect_device(self, device_path):

        device = dbus.Interface(
            self.bus.get_object(
                SERVICE_NAME,
                device_path),
            DEVICE_INTERFACE)
        try:
            device.Connect()
        except dbus.exceptions.DBusException as e:
            self.logger.exception(e)

    def remove_device(self, path):
        """Removes a device that's been either discovered, paired,
        connected, etc.

        :param path: The D-Bus path to the object
        :type path: string
        """

        self.adapter.RemoveDevice(
            self.bus.get_object(SERVICE_NAME, path))

    def find_device_by_address(self, address):
        """Finds the D-Bus path to a device that contains the
        specified address.

        :param address: The Bluetooth MAC address
        :type address: string
        :return: The path to the D-Bus object or None
        :rtype: string or None
        """

        # Find all connected/paired/discovered devices
        devices = find_objects(
            self.bus,
            SERVICE_NAME,
            DEVICE_INTERFACE)
        for path in devices:
            # Get the device's address and paired status
            device_props = dbus.Interface(
                self.bus.get_object(SERVICE_NAME, path),
                "org.freedesktop.DBus.Properties")
            device_addr = device_props.Get(
                DEVICE_INTERFACE,
                "Address").upper()

            # Check for an address match
            if device_addr != address.upper():
                continue
            return path

        return None
    
    def find_connected_devices(self, alias_filter=False):
        """Finds the D-Bus path to a device that contains the
        specified address.

        :param address: The Bluetooth MAC address
        :type address: string
        :return: The path to the D-Bus object or None
        :rtype: string or None
        """

        devices = find_objects(
            self.bus,
            SERVICE_NAME,
            DEVICE_INTERFACE)
        conn_devices = []
        for path in devices:
            # Get the device's connection status
            device_props = dbus.Interface(
                self.bus.get_object(SERVICE_NAME, path),
                "org.freedesktop.DBus.Properties")
            device_conn_status = device_props.Get(
                DEVICE_INTERFACE,
                "Connected")
            device_alias = device_props.Get(
                DEVICE_INTERFACE,
                "Alias").upper()

            if device_conn_status:
                if alias_filter and device_alias == alias_filter.upper():
                    conn_devices.append(path)
                else:
                    conn_devices.append(path)

        return conn_devices



from pathlib import Path

from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
        name='libnxctrl',
        version='0.1.6',
        url='https://github.com/Victrid/libnxctrl',
        license='GPLv3',
        author='Weihao Jiang',
        author_email='weihau.chiang@gmail.com',
        description='Python Library Emulating Nintendo Switch Controllers',
        long_description=long_description,
        long_description_content_type='text/markdown',
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Topic :: Software Development :: Libraries",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Natural Language :: English",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python :: 3",
            ],
        install_requires=[
            # NXBT requirements
            "dbus-python~=1.2.16",
            "Flask>=1.1.2,<2.3.0",
            "Flask-SocketIO>=5.0.1,<5.4.0",
            "eventlet>=0.31,<0.34",
            "blessed>=1.17.10,<1.20.0",
            "pynput~=1.7.1",
            "psutil~=5.6.6",
            "cryptography>=3.3.2,<37.1.0",
            # JoyControl requirements
            'hid~=1.0.5',
            'aioconsole~=0.5.1'
            ],
        packages=[
            "libnxctrl",
            "libnxctrl.nxbt.nxbt",
            "libnxctrl.nxbt.nxbt.controller",
            "libnxctrl.mart1no_joycontrol.joycontrol",
            "libnxctrl.Poohl_joycontrol.joycontrol"
            ],
        package_dir={
            'libnxctrl':                               'src/libnxctrl',
            'libnxctrl.nxbt.nxbt':                     'src/libnxctrl/nxbt/nxbt',
            'libnxctrl.mart1no_joycontrol.joycontrol': 'src/libnxctrl/mart1no_joycontrol/joycontrol',
            'libnxctrl.Poohl_joycontrol.joycontrol':   'src/libnxctrl/Poohl_joycontrol/joycontrol'
            },
        package_data={
            'libnxctrl.mart1no_joycontrol.joycontrol': [
                'profile/sdp_record_hid.xml',
                '../LICENSE',
                '../README.md'
                ],
            'libnxctrl.Poohl_joycontrol.joycontrol':   [
                'profile/sdp_record_hid.xml',
                '../LICENSE',
                '../README.md'
                ],
            'libnxctrl.nxbt.nxbt':                     [
                'controller/sdp/switch-controller.xml'
                '../LICENSE',
                '../README.md'
                ],
            'libnxctrl.nxbt.nxbt.controller':          [
                'sdp/switch-controller.xml'
                ]
            },
        )



from .controller import ControllerServer
from .controller import ControllerProtocol
from .controller import SwitchReportParser
from .controller import SwitchResponses
from .controller import Controller
from .bluez import *
from .nxbt import Nxbt
from .nxbt import Buttons
from .nxbt import Sticks
from .nxbt import JOYCON_L
from .nxbt import JOYCON_R
from .nxbt import PRO_CONTROLLER



import logging
import uuid
import dbus

from joycontrol import utils

logger = logging.getLogger(__name__)


HID_UUID = '00001124-0000-1000-8000-00805f9b34fb'
HID_PATH = '/bluez/switch/hid'


class HidDevice:
    def __init__(self, device_id=None):
        self._device_id = device_id
        bus = dbus.SystemBus()
        # Get Bluetooth adapter from dbus interface
        for path, ifaces in bus.get_object('org.bluez', '/').GetManagedObjects(dbus_interface='org.freedesktop.DBus.ObjectManager').items():
            adapter_info = ifaces.get('org.bluez.Adapter1')
            if adapter_info and (device_id is None or device_id == adapter_info['Address'] or path.endswith(str(device_id))):
                self.dev = bus.get_object('org.bluez', path)
                break
        else:
            raise ValueError(f'Adapter {device_id} not found.')

        self.adapter = dbus.Interface(self.dev, 'org.bluez.Adapter1')
        # The sad news is someone decided that this convoluted mess passing
        # strings back and forth to get properties would be simpler than literal
        # adapter.some_property = 4 or adapter.some_property_set(4)
        self.properties = dbus.Interface(self.dev, 'org.freedesktop.DBus.Properties')
        self._adapter_name = self.dev.object_path.split("/")[-1]

    def get_address(self) -> str:
        """
        :returns adapter Bluetooth address
        """
        return str(self.properties.Get(self.adapter.dbus_interface, "Address"))

    async def set_address(self, bt_addr, interactive=True):
        if not interactive:
            return False
        # TODO: automated detection
        print(f"Attempting to change the bluetooth MAC to {bt_addr}")
        print("please choose your method:")
        print("\t1: bdaddr - ericson, csr, TI, broadcom, zeevo, st")
        print("\t2: hcitool - intel chipsets")
        print("\t3: hcitool - cypress (raspberri pi 3B+ & 4B)")
        print("\tx: abort, dont't change")
        hci_version = " ".join(reversed(list(map(lambda h: '0x' + h, bt_addr.split(":")))))
        c = input()
        if c == '1':
            await utils.run_system_command(f'bdaddr -i {self._adapter_name} {bt_addr}')
        elif c == '2':
            await utils.run_system_command(f'hcitool cmd 0x3f 0x0031 {hci_version}')
        elif c == '3':
            await utils.run_system_command(f'hcitool cmd 0x3f 0x001 {hci_version}')
        else:
            return False
        await utils.run_system_command("hciconfig hci0 reset")
        await utils.run_system_command("systemctl restart bluetooth.service")

        # now we have to reget all dbus-shenanigans because we just restarted it's service.
        self.__init__(self._device_id)

        if self.get_address() != bt_addr:
            logger.info("Failed to set btaddr")
            return False
        else:
            logger.info(f"Changed bt_addr to {bt_addr}")
            return True

    def get_paired_switches(self):
        switches = []
        for path, ifaces in dbus.SystemBus().get_object('org.bluez', '/').GetManagedObjects('org.freedesktop.DBus.ObjectManager', dbus_interface='org.freedesktop.DBus.ObjectManager').items():
            d = ifaces.get("org.bluez.Device1")
            if d and d['Name'] == "Nintendo Switch":
                switches += [path]
        return switches

    def unpair_path(self, path):
        self.adapter.RemoveDevice(path)

    def powered(self, boolean=True):
        self.properties.Set(self.adapter.dbus_interface, 'Powered', boolean)

    def discoverable(self, boolean=True):
        """
        Make adapter discoverable, starts advertising.
        """
        self.properties.Set(self.adapter.dbus_interface, 'Discoverable', boolean)

    def pairable(self, boolean=True):
        """
        Make adapter pairable
        """
        self.properties.Set(self.adapter.dbus_interface, 'Pairable', boolean)

    async def set_class(self, cls='0x002508'):
        """
        Sets Bluetooth device class. Requires hciconfig system command.
        :param cls: default 0x002508 (Gamepad/joystick device class)
        """
        logger.info(f'setting device class to {cls}...')
        await utils.run_system_command(f'hciconfig {self._adapter_name} class {cls}')
        if self.properties.Get(self.adapter.dbus_interface, "Class") != int(cls, base=0):
            logger.error(f"Could not set class to the required {cls}. Connecting probably won't work.")

    async def set_name(self, name: str):
        """
        Set Bluetooth device name.
        :param name: to set.
        """
        logger.info(f'setting device name to {name}...')
        self.properties.Set(self.adapter.dbus_interface, 'Alias', name)

    def get_UUIDs(self):
        return self.properties.Get(self.adapter.dbus_interface, "UUIDs")

    @staticmethod
    def register_sdp_record(record_path):
        _uuid = str(uuid.uuid4())

        with open(record_path) as record:
            opts = {
                'ServiceRecord': record.read(),
                'Role': 'server',
                'Service': HID_UUID,
                'RequireAuthentication': False,
                'RequireAuthorization': False
            }
            bus = dbus.SystemBus()
            manager = dbus.Interface(bus.get_object("org.bluez", "/org/bluez"), "org.bluez.ProfileManager1")
            manager.RegisterProfile(HID_PATH, _uuid, opts)

    @staticmethod
    def get_address_of_paired_path(path):
        return str(dbus.SystemBus().get_object('org.bluez', path).Get('org.bluez.Device1', "Address", dbus_interface='org.freedesktop.DBus.Properties'))



import logging
import datetime


def configure(console_level=logging.DEBUG, file_level=logging.DEBUG, logfile_name=None):
    """
    Configures logging formatting

    :param console_level: log level of console logger
    :param file_level: log lever of file logger
    :param logfile_name: name of logfile
    """
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter(
        "[%(asctime)s] %(name)s %(funcName)s::%(lineno)s %(levelname)s - %(message)s",
        "%H:%M:%S"
    )

    # create console logger
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)
    console_handler.setLevel(console_level)

    root_logger.addHandler(console_handler)

    # create file logger
    if logfile_name is not None:
        today = datetime.datetime.now()
        name_of_file = today.strftime(f'%Y-%m-%d_%H-%M_{logfile_name}.log')

        file_handler = logging.FileHandler(name_of_file)
        file_handler.setLevel(file_level)
        file_handler.setFormatter(formatter)

        root_logger.addHandler(file_handler)


if __name__ == "__main__":
    # Run test output on stdout
    configure()

    logger = logging.getLogger("test")

    def test():
        logger.debug("debug msg")
        logger.info("info msg")
        logger.warning("warning msg")

    def test2():
        logger.error("error msg")
        logger.critical("critical msg")

    # test debug, info, warning
    test()
    # test error, critical
    test2()

    # test exceptions
    try:
        raise RuntimeError("It's a trap!")
    except Exception as e:
        logger.exception(e)



import inspect
import logging
import shlex

from aioconsole import ainput

from joycontrol.controller_state import button_push, ControllerState
from joycontrol.transport import NotConnectedError

logger = logging.getLogger(__name__)


def _print_doc(string):
    """
    Attempts to remove common white space at the start of the lines in a doc string
    to unify the output of doc strings with different indention levels.

    Keeps whitespace lines intact.

    :param fun: function to print the doc string of
    """
    lines = string.split('\n')
    if lines:
        prefix_i = 0
        for i, line_0 in enumerate(lines):
            # find non empty start lines
            if line_0.strip():
                # traverse line and stop if character mismatch with other non empty lines
                for prefix_i, c in enumerate(line_0):
                    if not c.isspace():
                        break
                    if any(lines[j].strip() and (prefix_i >= len(lines[j]) or c != lines[j][prefix_i])
                           for j in range(i+1, len(lines))):
                        break
                break

        for line in lines:
            print(line[prefix_i:] if line.strip() else line)


class CLI:
    def __init__(self):
        self.commands = {}

    def add_command(self, name, command):
        if name in self.commands:
            raise ValueError(f'Command {name} already registered.')
        self.commands[name] = command

    async def cmd_help(self):
        print('Commands:')
        for name, fun in inspect.getmembers(self):
            if name.startswith('cmd_') and fun.__doc__:
                _print_doc(fun.__doc__)

        for name, fun in self.commands.items():
            if fun.__doc__:
                _print_doc(fun.__doc__)

        print('Commands can be chained using "&&"')
        print('Type "exit" to close.')

    async def run(self):
        while True:
            user_input = await ainput(prompt='cmd >> ')
            if not user_input:
                continue

            for command in user_input.split('&&'):
                cmd, *args = shlex.split(command)

                if cmd == 'exit':
                    return

                if hasattr(self, f'cmd_{cmd}'):
                    try:
                        result = await getattr(self, f'cmd_{cmd}')(*args)
                        if result:
                            print(result)
                    except Exception as e:
                        print(e)
                elif cmd in self.commands:
                    try:
                        result = await self.commands[cmd](*args)
                        if result:
                            print(result)
                    except Exception as e:
                        print(e)
                else:
                    print('command', cmd, 'not found, call help for help.')

    @staticmethod
    def deprecated(message):
        async def dep_printer(*args, **kwargs):
            print(message)

        return dep_printer


class ControllerCLI(CLI):
    def __init__(self, controller_state: ControllerState):
        super().__init__()
        self.controller_state = controller_state

    async def cmd_help(self):
        print('Button commands:')
        print(', '.join(self.controller_state.button_state.get_available_buttons()))
        print()
        await super().cmd_help()

    @staticmethod
    def _set_stick(stick, direction, value):
        if direction == 'center':
            stick.set_center()
        elif direction == 'up':
            stick.set_up()
        elif direction == 'down':
            stick.set_down()
        elif direction == 'left':
            stick.set_left()
        elif direction == 'right':
            stick.set_right()
        elif direction in ('h', 'horizontal'):
            if value is None:
                raise ValueError(f'Missing value')
            try:
                val = int(value)
            except ValueError:
                raise ValueError(f'Unexpected stick value "{value}"')
            stick.set_h(val)
        elif direction in ('v', 'vertical'):
            if value is None:
                raise ValueError(f'Missing value')
            try:
                val = int(value)
            except ValueError:
                raise ValueError(f'Unexpected stick value "{value}"')
            stick.set_v(val)
        else:
            raise ValueError(f'Unexpected argument "{direction}"')

        return f'{stick.__class__.__name__} was set to ({stick.get_h()}, {stick.get_v()}).'

    async def cmd_stick(self, side, direction, value=None):
        """
        stick - Command to set stick positions.
        :param side: 'l', 'left' for left control stick; 'r', 'right' for right control stick
        :param direction: 'center', 'up', 'down', 'left', 'right';
                          'h', 'horizontal' or 'v', 'vertical' to set the value directly to the "value" argument
        :param value: horizontal or vertical value
        """
        if side in ('l', 'left'):
            stick = self.controller_state.l_stick_state
            return ControllerCLI._set_stick(stick, direction, value)
        elif side in ('r', 'right'):
            stick = self.controller_state.r_stick_state
            return ControllerCLI._set_stick(stick, direction, value)
        else:
            raise ValueError('Value of side must be "l", "left" or "r", "right"')

    async def run(self):
        while True:
            user_input = await ainput(prompt='cmd >> ')
            if not user_input:
                continue

            buttons_to_push = []

            for command in user_input.split('&&'):
                cmd, *args = shlex.split(command)

                if cmd == 'exit':
                    return

                available_buttons = self.controller_state.button_state.get_available_buttons()

                if hasattr(self, f'cmd_{cmd}'):
                    try:
                        result = await getattr(self, f'cmd_{cmd}')(*args)
                        if result:
                            print(result)
                    except Exception as e:
                        print(e)
                elif cmd in self.commands:
                    try:
                        result = await self.commands[cmd](*args)
                        if result:
                            print(result)
                    except Exception as e:
                        print(e)
                elif cmd in available_buttons:
                    buttons_to_push.append(cmd)
                else:
                    print('command', cmd, 'not found, call help for help.')

            if buttons_to_push:
                await button_push(self.controller_state, *buttons_to_push)
            else:
                try:
                    await self.controller_state.send()
                except NotConnectedError:
                    logger.info('Connection was lost.')
                    return



import asyncio

from joycontrol import utils
from joycontrol.controller import Controller
from joycontrol.memory import FlashMemory


class ControllerState:
    def __init__(self, protocol, controller: Controller, spi_flash: FlashMemory = None):
        self._protocol = protocol
        self._controller = controller
        self._nfc_content = None

        self._spi_flash = spi_flash

        self.button_state = ButtonState(controller)

        # create left stick state
        self.l_stick_state = self.r_stick_state = None
        if controller in (Controller.PRO_CONTROLLER, Controller.JOYCON_L):
            # load calibration data from memory
            calibration = None
            if spi_flash is not None:
                calibration_data = spi_flash.get_user_l_stick_calibration()
                if calibration_data is None:
                    calibration_data = spi_flash.get_factory_l_stick_calibration()
                calibration = LeftStickCalibration.from_bytes(calibration_data)

            self.l_stick_state = StickState(calibration=calibration)
            if calibration is not None:
                self.l_stick_state.set_center()

        # create right stick state
        if controller in (Controller.PRO_CONTROLLER, Controller.JOYCON_R):
            # load calibration data from memory
            calibration = None
            if spi_flash is not None:
                calibration_data = spi_flash.get_user_r_stick_calibration()
                if calibration_data is None:
                    calibration_data = spi_flash.get_factory_r_stick_calibration()
                calibration = RightStickCalibration.from_bytes(calibration_data)

            self.r_stick_state = StickState(calibration=calibration)
            if calibration is not None:
                self.r_stick_state.set_center()

        self.sig_is_send = asyncio.Event()

    def get_controller(self):
        return self._controller

    def get_flash_memory(self):
        return self._spi_flash

    def set_nfc(self, nfc_content):
        self._nfc_content = nfc_content

    def get_nfc(self):
        return self._nfc_content

    async def send(self):
        """
        Invokes protocol.send_controller_state(). Returns after the controller state was send.
        Raises NotConnected exception if the connection was lost.
        """
        await self._protocol.send_controller_state()

    async def connect(self):
        """
        Waits until the switch is paired with the controller and accepts button commands
        """
        await self._protocol.sig_set_player_lights.wait()


class ButtonState:
    """
    Utility class to set buttons in the input report
    https://github.com/dekuNukem/Nintendo_Switch_Reverse_Engineering/blob/master/bluetooth_hid_notes.md
    Byte 	0 	    1 	    2 	    3 	    4 	    5 	    6 	    7
    1   	Y 	    X 	    B 	    A 	    SR 	    SL 	    R 	    ZR
    2       Minus 	Plus 	R Stick L Stick Home 	Capture
    3       Down 	Up 	    Right 	Left 	SR 	    SL 	    L 	    ZL

    Example for generated methods: home button (byte_2, 4)

    def home(self, pushed=True):
        if pushed != utils.get_bit(self.byte_2, 4):
            self.byte_2 = utils.flip_bit(self.byte_2, 4)

    def home_is_set(self):
        return get_bit(self.byte_2, 4)
    """
    def __init__(self, controller: Controller):
        self.controller = controller

        # 3 bytes
        self._byte_1 = 0
        self._byte_2 = 0
        self._byte_3 = 0

        # generating methods for each button
        def button_method_factory(byte, bit):
            def setter(pushed=True):
                _byte = getattr(self, byte)

                if pushed != utils.get_bit(_byte, bit):
                    setattr(self, byte, utils.flip_bit(_byte, bit))

            def getter():
                return utils.get_bit(getattr(self, byte), bit)
            return setter, getter

        if self.controller == Controller.PRO_CONTROLLER:
            self._available_buttons = {'y', 'x', 'b', 'a', 'r', 'zr',
                                       'minus', 'plus', 'r_stick', 'l_stick', 'home', 'capture',
                                       'down', 'up', 'right', 'left', 'l', 'zl'}
        elif self.controller == Controller.JOYCON_R:
            self._available_buttons = {'y', 'x', 'b', 'a', 'sr', 'sl', 'r', 'zr',
                                       'plus', 'r_stick', 'home'}
        elif self.controller == Controller.JOYCON_L:
            self._available_buttons = {'minus', 'l_stick', 'capture',
                                       'down', 'up', 'right', 'left', 'sr', 'sl', 'l', 'zl'}

        # byte 1
        if self.controller == Controller.PRO_CONTROLLER or self.controller == Controller.JOYCON_R:
            self.y, self.y_is_set = button_method_factory('_byte_1', 0)
            self.x, self.x_is_set = button_method_factory('_byte_1', 1)
            self.b, self.b_is_set = button_method_factory('_byte_1', 2)
            self.a, self.a_is_set = button_method_factory('_byte_1', 3)

            if self.controller == Controller.JOYCON_R:
                self.sr, self.sr_is_set = button_method_factory('_byte_1', 4)
                self.sl, self.sl_is_set = button_method_factory('_byte_1', 5)

            self.r, self.r_is_set = button_method_factory('_byte_1', 6)
            self.zr, self.zr_is_set = button_method_factory('_byte_1', 7)

        # byte 2
        self.minus, self.minus_is_set = button_method_factory('_byte_2', 0)
        self.plus, self.plus_is_set = button_method_factory('_byte_2', 1)
        self.r_stick, self.r_stick_is_set = button_method_factory('_byte_2', 2)
        self.l_stick, self.l_stick_is_set = button_method_factory('_byte_2', 3)
        if self.controller == Controller.JOYCON_R or self.controller == Controller.PRO_CONTROLLER:
            self.home, self.home_is_set = button_method_factory('_byte_2', 4)
        if self.controller == Controller.JOYCON_L or self.controller == Controller.PRO_CONTROLLER:
            self.capture, self.capture_is_set = button_method_factory('_byte_2', 5)

        # byte 3
        if self.controller == Controller.PRO_CONTROLLER or self.controller == Controller.JOYCON_L:
            self.down, self.down_is_set = button_method_factory('_byte_3', 0)
            self.up, self.up_is_set = button_method_factory('_byte_3', 1)
            self.right, self.right_is_set = button_method_factory('_byte_3', 2)
            self.left, self.left_is_set = button_method_factory('_byte_3', 3)

            if self.controller == Controller.JOYCON_L:
                self.sr, self.sr_is_set = button_method_factory('_byte_3', 4)
                self.sl, self.sl_is_set = button_method_factory('_byte_3', 5)

            self.l, self.l_is_set = button_method_factory('_byte_3', 6)
            self.zl, self.zl_is_set = button_method_factory('_byte_3', 7)

    def set_button(self, button, pushed=True):
        if button not in self._available_buttons:
            raise ValueError(f'Given button "{button}" is not available to {self.controller.device_name()}.')
        getattr(self, button)(pushed=pushed)

    def get_button(self, button):
        if button not in self._available_buttons:
            raise ValueError(f'Given button "{button}" is not available to {self.controller.device_name()}.')
        return getattr(self, f'{button}_is_set')()

    def get_available_buttons(self):
        """
        :returns: set of valid buttons
        """
        return set(self._available_buttons)

    def __iter__(self):
        """
        :returns: iterator over the button bytes
        """
        yield self._byte_1
        yield self._byte_2
        yield self._byte_3

    def clear(self):
        self._byte_1 = self._byte_2 = self._byte_3 = 0


async def button_press(controller_state, *buttons):
    """
    Set given buttons in the controller state to the pressed down state and wait till send.
    :param controller_state:
    :param buttons: Buttons to press down (see ButtonState.get_available_buttons)
    """
    if not buttons:
        raise ValueError('No Buttons were given.')

    button_state = controller_state.button_state

    for button in buttons:
        # push button
        button_state.set_button(button, pushed=True)

    # wait until report is send
    await controller_state.send()


async def button_release(controller_state, *buttons):
    """
    Set given buttons in the controller state to the unpressed state and wait till send.
    :param controller_state:
    :param buttons: Buttons to set to unpressed (see ButtonState.get_available_buttons)
    """
    if not buttons:
        raise ValueError('No Buttons were given.')

    button_state = controller_state.button_state

    for button in buttons:
        # release button
        button_state.set_button(button, pushed=False)

    # wait until report is send
    await controller_state.send()


async def button_push(controller_state, *buttons, sec=0.1):
    """
    Shortly push the given buttons. Wait until the controller state is send.
    :param controller_state:
    :param buttons: Buttons to push (see ButtonState.get_available_buttons)
    :param sec: Seconds to wait before releasing the button, default: 0.1
    """
    await button_press(controller_state, *buttons)
    await asyncio.sleep(sec)
    await button_release(controller_state, *buttons)


class _StickCalibration:
    def __init__(self, h_center, v_center, h_max_above_center, v_max_above_center, h_max_below_center, v_max_below_center):
        self.h_center = h_center
        self.v_center = v_center

        self.h_max_above_center = h_max_above_center
        self.v_max_above_center = v_max_above_center
        self.h_max_below_center = h_max_below_center
        self.v_max_below_center = v_max_below_center

    def __str__(self):
        return f'h_center:{self.h_center} v_center:{self.v_center} h_max_above_center:{self.h_max_above_center} ' \
               f'v_max_above_center:{self.v_max_above_center} h_max_below_center:{self.h_max_below_center} ' \
               f'v_max_below_center:{self.v_max_below_center}'


class LeftStickCalibration(_StickCalibration):
    @staticmethod
    def from_bytes(_9bytes):
        h_max_above_center = (_9bytes[1] << 8) & 0xF00 | _9bytes[0]
        v_max_above_center = (_9bytes[2] << 4) | (_9bytes[1] >> 4)
        h_center =           (_9bytes[4] << 8) & 0xF00 | _9bytes[3]
        v_center =           (_9bytes[5] << 4) | (_9bytes[4] >> 4)
        h_max_below_center = (_9bytes[7] << 8) & 0xF00 | _9bytes[6]
        v_max_below_center = (_9bytes[8] << 4) | (_9bytes[7] >> 4)

        return _StickCalibration(h_center, v_center, h_max_above_center, v_max_above_center,
                                 h_max_below_center, v_max_below_center)


class RightStickCalibration(_StickCalibration):
    @staticmethod
    def from_bytes(_9bytes):
        h_center =           (_9bytes[1] << 8) & 0xF00 | _9bytes[0]
        v_center =           (_9bytes[2] << 4) | (_9bytes[1] >> 4)
        h_max_below_center = (_9bytes[4] << 8) & 0xF00 | _9bytes[3]
        v_max_below_center = (_9bytes[5] << 4) | (_9bytes[4] >> 4)
        h_max_above_center = (_9bytes[7] << 8) & 0xF00 | _9bytes[6]
        v_max_above_center = (_9bytes[8] << 4) | (_9bytes[7] >> 4)

        return _StickCalibration(h_center, v_center, h_max_above_center, v_max_above_center,
                                 h_max_below_center, v_max_below_center)


class StickState:
    def __init__(self, h=0, v=0, calibration: _StickCalibration = None):
        for val in (h, v):
            if not 0 <= val < 0x1000:
                raise ValueError(f'Stick values must be in [0,{0x1000})')

        self._h_stick = h
        self._v_stick = v

        self._calibration = calibration

    def set_h(self, value):
        if not 0 <= value < 0x1000:
            raise ValueError(f'Stick values must be in [0,{0x1000})')
        self._h_stick = value

    def get_h(self):
        return self._h_stick

    def set_v(self, value):
        if not 0 <= value < 0x1000:
            raise ValueError(f'Stick values must be in [0,{0x1000})')
        self._v_stick = value

    def get_v(self):
        return self._v_stick

    def set_center(self):
        """
        Sets stick to center position using the calibration data.
        """
        if self._calibration is None:
            raise ValueError('No calibration data available.')
        self._h_stick = self._calibration.h_center
        self._v_stick = self._calibration.v_center

    def is_center(self, radius=0):
        return self._calibration.h_center - radius <= self._h_stick <= self._calibration.h_center + radius and \
               self._calibration.v_center - radius <= self._v_stick <= self._calibration.v_center + radius

    def set_up(self):
        """
        Sets stick to up position using the calibration data.
        """
        if self._calibration is None:
            raise ValueError('No calibration data available.')
        self._h_stick = self._calibration.h_center
        self._v_stick = self._calibration.v_center + self._calibration.v_max_above_center

    def set_down(self):
        """
        Sets stick to down position using the calibration data.
        """
        if self._calibration is None:
            raise ValueError('No calibration data available.')
        self._h_stick = self._calibration.h_center
        self._v_stick = self._calibration.v_center - self._calibration.v_max_below_center

    def set_left(self):
        """
        Sets stick to left position using the calibration data.
        """
        if self._calibration is None:
            raise ValueError('No calibration data available.')
        self._h_stick = self._calibration.h_center - self._calibration.h_max_below_center
        self._v_stick = self._calibration.v_center

    def set_right(self):
        """
        Sets stick to right position using the calibration data.
        """
        if self._calibration is None:
            raise ValueError('No calibration data available.')
        self._h_stick = self._calibration.h_center + self._calibration.h_max_above_center
        self._v_stick = self._calibration.v_center

    def set_calibration(self, calibration):
        self._calibration = calibration

    def get_calibration(self):
        if self._calibration is None:
            raise ValueError('No calibration data available.')
        return self._calibration

    @staticmethod
    def from_bytes(_3bytes):
        stick_h = _3bytes[0] | ((_3bytes[1] & 0xF) << 8)
        stick_v = (_3bytes[1] >> 4) | (_3bytes[2] << 4)

        return StickState(h=stick_h, v=stick_v)

    def __bytes__(self):
        byte_1 = 0xFF & self._h_stick
        byte_2 = (self._h_stick >> 8) | ((0xF & self._v_stick) << 4)
        byte_3 = self._v_stick >> 4
        assert all(0 <= byte <= 0xFF for byte in (byte_1, byte_2, byte_3))
        return bytes((byte_1, byte_2, byte_3))



from typing import Type

from .wrapper import NXWrapper


def get_backend(backend_name: str) -> Type[NXWrapper]:
    if backend_name == "nxbt":
        from .nxbt_wrapper import NXBTControl
        return NXBTControl
    elif backend_name == "mart1no":
        from .joycontrol_wrapper import Mart1noJoyControl
        raise NotImplementedError("Joycontrol support is not implemented.")
        # return Mart1noJoyControl
    elif backend_name == "poohl":
        from .joycontrol_wrapper import PoohlJoyControl
        raise NotImplementedError("Joycontrol support is not implemented.")
        # return PoohlJoyControl
    else:
        raise ValueError("Unknown backend: {}".format(backend_name))
